function interrupt_cutscene()
    ---@type MarioState
    local m = gMarioStates[0]
    ---@type Camera
    local c = gMarioStates[0].area.camera

    if not m or not c then
        return
    end

    if IS_VANILLA and (gGlobalSyncTable.darkStarFlags == 7 or _G.bossRushDarkStar) and c.cutscene == CUTSCENE_STAR_SPAWN then
        local grand = obj_get_first_with_behavior_id(id_bhvGrandStar)
        if grand then
            local m2 = nearest_mario_state_to_object(grand)
            if m2.playerIndex == 0 then
                obj_mark_for_deletion(grand)
                network_send_object(grand, true)
                spawn_sync_object(id_bhvGrandStarBoss, E_MODEL_STAR, grand.oPosX + grand.oVelX, grand.oPosY + grand
                    .oVelY, grand.oPosZ + grand.oVelZ, function(o)
                        o.oFaceAngleYaw = grand.oFaceAngleYaw
                        o.oFaceAnglePitch = grand.oFaceAnglePitch
                        o.oFaceAngleRoll = grand.oFaceAngleRoll
                        o.oAngleVelYaw = grand.oAngleVelYaw
                    end)
            end
            c.cutscene = 0
            play_cutscene(c)
        end
    end
end

hook_event(HOOK_UPDATE, interrupt_cutscene)

local GRAND_BOSS_ACT_INTRO = 0
local GRAND_BOSS_ACT_IDLE = 1
local GRAND_BOSS_ACT_TORNADO = 2
local GRAND_BOSS_ACT_POUND = 3
local GRAND_BOSS_ACT_SUMMON_STARS = 4
local GRAND_BOSS_ACT_HURT = 5
local GRAND_BOSS_ACT_DIVE = 6
local GRAND_BOSS_ACT_SLAM = 7
local GRAND_BOSS_ACT_DYING = 8

local playingFinalBoss = false

gGlobalSyncTable.secretBoss = false

---@param o Object
function grand_star_boss_init(o)
    o.oFlags = o.oFlags |
        (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
    cur_obj_scale(2)

    local sGrandBossHitbox = get_temp_object_hitbox()
    sGrandBossHitbox.interactType = INTERACT_DAMAGE
    sGrandBossHitbox.downOffset = 50
    sGrandBossHitbox.damageOrCoinValue = 2
    sGrandBossHitbox.health = 3
    sGrandBossHitbox.numLootCoins = 0
    sGrandBossHitbox.radius = 80
    sGrandBossHitbox.height = 100
    sGrandBossHitbox.hurtboxRadius = 80
    sGrandBossHitbox.hurtboxHeight = 100
    obj_set_hitbox(o, sGrandBossHitbox)

    o.oIntangibleTimer = -1
    o.oBobombBlinkTimer = 0
    o.oHomeX, o.oHomeY, o.oHomeZ = 0, 500, 0
    o.oFloorHeight = 307
    o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 40, 0, -0.5, 10, 10, 0
    playingFinalBoss = false
    network_init_object(o, true, { 'oBobombBlinkTimer', 'oAngleVelYaw', 'oAngleVelPitch', 'oHealth' })
end

---@param o Object
function grand_star_boss_loop(o)
    if (not playingFinalBoss) and o.oAction ~= GRAND_BOSS_ACT_INTRO and o.oAction ~= GRAND_BOSS_ACT_DYING then
        obj_set_model_extended(o, E_MODEL_DARK_STAR)
        set_background_music(0, secert_boss_music, 10)
        obj_set_secondary_camera_focus()
        playingFinalBoss = true
        o.oIntangibleTimer = 0
    end

    local m = nearest_mario_state_to_object(o)
    local player = m.marioObj
    local distanceToPlayer = lateral_dist_between_objects(o, player) or 10000
    local angleToPlayer = obj_angle_to_object(o, player) or 0

    if o.oAction == GRAND_BOSS_ACT_INTRO then
        if (o.oAngleVelYaw < 0x2000) then
            o.oAngleVelYaw = o.oAngleVelYaw + 0x100
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + o.oAngleVelYaw
        if o.oTimer % 5 == 0 then
            cur_obj_play_sound_2(SOUND_GENERAL_GRAND_STAR_JUMP)
        elseif o.oTimer > 90 then
            cur_obj_change_action(GRAND_BOSS_ACT_TORNADO)
            o.oBobombBlinkTimer = 1
        end
    elseif o.oAction == GRAND_BOSS_ACT_IDLE then
        obj_turn_pitch_toward_mario(m, 120, 2000)
        cur_obj_rotate_yaw_toward(angleToPlayer, 2000);

        if o.oPosY > o.oHomeY + 16 then
            o.oPosY = o.oPosY - (o.oPosY - o.oHomeY + 20) // 10
        else
            o.oPosY = o.oHomeY + 16 * coss(4000 * o.oTimer)
        end

        o.oFaceAngleYaw = o.oMoveAngleYaw
        o.oFaceAnglePitch = o.oMoveAnglePitch
        if o.oTimer > 90 then
            o.oBobombBlinkTimer = o.oBobombBlinkTimer + 1
            if o.oBobombBlinkTimer == 6 - o.oHealth then
                cur_obj_change_action(GRAND_BOSS_ACT_SUMMON_STARS)
                o.oBobombBlinkTimer = 1
            elseif o.oBobombBlinkTimer == 2 then
                cur_obj_change_action(GRAND_BOSS_ACT_POUND)
                cur_obj_play_sound_2(SOUND_GENERAL_GRAND_STAR_JUMP)
                o.oVelY = 40
            elseif o.oBobombBlinkTimer == 3 then
                cur_obj_change_action(GRAND_BOSS_ACT_DIVE)
            elseif o.oBobombBlinkTimer == 4 then
                cur_obj_change_action(GRAND_BOSS_ACT_SLAM)
            else
                cur_obj_change_action(GRAND_BOSS_ACT_TORNADO)
            end
        end
    elseif o.oAction == GRAND_BOSS_ACT_TORNADO then
        o.oFaceAnglePitch = 0
        o.oMoveAnglePitch = 0
        o.oVelY = 0
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x2000
        obj_forward_vel_approach(10, 2)
        cur_obj_play_sound_2(SOUND_ENV_WIND1)

        local oldMoveYaw = o.oMoveAngleYaw
        o.oMoveAngleYaw = o.oTimer * 0x600
        cur_obj_spawn_strong_wind_particles(1, 3, 0, o.oFloorHeight - o.oPosY - 50, 0)
        o.oMoveAngleYaw = oldMoveYaw

        cur_obj_rotate_yaw_toward(angleToPlayer, 800)
        cur_obj_move_using_fvel_and_gravity()
        if o.oTimer > 180 then
            cur_obj_change_action(GRAND_BOSS_ACT_IDLE)
        end
    elseif o.oAction == GRAND_BOSS_ACT_POUND then
        o.oFaceAnglePitch = 0
        o.oMoveAnglePitch = 0
        obj_forward_vel_approach(15, 2)

        cur_obj_rotate_yaw_toward(angleToPlayer, 800)
        o.oFaceAngleYaw = o.oMoveAngleYaw
        o.oGravity = -2
        cur_obj_move_using_fvel_and_gravity()
        o.oGravity = 0

        if o.oFloorHeight >= o.oPosY and o.oVelY <= 0 then
            o.oVelY = 60
            o.oPosY = o.oFloorHeight
            cur_obj_play_sound_2(SOUND_ACTION_TERRAIN_HEAVY_LANDING)
            set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)
            spawn_non_sync_object(
                id_bhvBowserShockWave,
                E_MODEL_BOWSER_WAVE,
                o.oPosX, o.oPosY, o.oPosZ,
                nil
            )

            for i = 0, (MAX_PLAYERS - 1) do
                local m2 = gMarioStates[i]
                if is_player_active(m2) ~= 0 and o.oFloorHeight == m2.pos.y and m2.invincTimer == 0 and (m2.action & ACT_FLAG_INVULNERABLE) == 0 and m2.health > 0xFF then
                    o.oInteractionSubtype = o.oInteractionSubtype & ~INT_SUBTYPE_DELAY_INVINCIBILITY
                    take_damage_and_knock_back(m2, o)
                end
            end

            if o.oTimer > 210 then
                cur_obj_change_action(GRAND_BOSS_ACT_IDLE)
            end
        end
    elseif o.oAction == GRAND_BOSS_ACT_SUMMON_STARS then
        o.oFaceAnglePitch = 0
        o.oMoveAnglePitch = 0
        o.oVelY = 0
        local homeDist = cur_obj_lateral_dist_to_home()
        o.oMoveAngleYaw = cur_obj_angle_to_home()
        if homeDist < 100 then
            o.oForwardVel = 0
        else
            o.oForwardVel = clamp(1, homeDist // 10, 50)
            cur_obj_move_using_fvel_and_gravity()
        end

        if (o.oAngleVelYaw < 0x2000) then
            o.oAngleVelYaw = o.oAngleVelYaw + 0x100
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + o.oAngleVelYaw
        if o.oTimer % 15 == 0 then
            cur_obj_play_sound_2(SOUND_GENERAL_GRAND_STAR_JUMP)
            if m.playerIndex == 0 then
                local x = math.random(-1200, 1200)
                local y = o.oFloorHeight + 30 * 60
                local z = math.random(-1200, 1200)
                spawn_sync_object(id_bhvFallingStar, E_MODEL_DARK_STAR, x, y, z - 15 * 60, function(star)
                    star.parentObj = o
                    star.oHomeX = x
                    star.oHomeY = o.oFloorHeight
                    star.oHomeZ = z
                    star.oBehParams = math.random(0, 4 - o.oHealth)
                end)
            end
        elseif o.oTimer > 250 then
            cur_obj_change_action(GRAND_BOSS_ACT_IDLE)
        end
    elseif o.oAction == GRAND_BOSS_ACT_HURT then
        if o.oTimer == 1 then
            cur_obj_play_sound_2(SOUND_OBJ2_WHOMP_SOUND_SHORT)
            o.oHealth = o.oHealth - 1
            o.oForwardVel = -80
            o.oVelY = 100
            o.oAngleVelPitch = -0x1200
        end
        o.oFaceAnglePitch = o.oFaceAnglePitch + o.oAngleVelPitch
        if o.oAngleVelPitch < 0 then
            o.oAngleVelPitch = o.oAngleVelPitch + 0x50
        else
            o.oFaceAnglePitch = 0
        end

        if o.oForwardVel < 0 then
            o.oForwardVel = o.oForwardVel + 2
        elseif o.oTimer > 90 then
            if o.oHealth <= 0 then
                cur_obj_change_action(GRAND_BOSS_ACT_DYING)
            else
                cur_obj_change_action(GRAND_BOSS_ACT_IDLE)
            end
            o.oBobombBlinkTimer = 0
        end

        if o.oVelY > 0 then
            o.oGravity = -2
        end
        cur_obj_move_using_fvel_and_gravity()
        o.oGravity = 0
    elseif o.oAction == GRAND_BOSS_ACT_DIVE then
        o.oFaceAnglePitch = (o.oForwardVel * 0x3000) // 200
        o.oMoveAnglePitch = 0
        o.oVelY = 0
        if o.oSubAction == 0 then
            local homeDist = dist_between_object_and_point(o, o.oHomeX + 5000, o.oHomeY, o.oHomeZ)
            if homeDist < 100 then
                o.oForwardVel = 0
                cur_obj_rotate_yaw_toward(angleToPlayer, 4000)
                if o.oTimer > 120 then
                    cur_obj_play_sound_2(SOUND_GENERAL_GRAND_STAR)
                    o.oSubAction = 1
                end
            else
                o.oMoveAngleYaw = obj_angle_to_point(o, o.oHomeX + 5000, o.oHomeZ)
                o.oForwardVel = clamp(1, homeDist // 10, 50)
                cur_obj_move_using_fvel_and_gravity()
            end
            o.oFaceAngleYaw = o.oMoveAngleYaw
        elseif o.oSubAction == 1 then
            if o.oForwardVel < 200 then
                o.oForwardVel = o.oForwardVel + 20
            end
            cur_obj_rotate_yaw_toward(angleToPlayer, 4000)
            o.oFaceAngleYaw = o.oMoveAngleYaw
            o.oPosY = o.oFloorHeight + 160
            cur_obj_move_using_fvel_and_gravity()
            if distanceToPlayer < 150 then
                o.oSubAction = 2
                spawn_non_sync_object(
                    id_bhvFlameLargeBurningOut,
                    E_MODEL_RED_FLAME,
                    o.oPosX, o.oFloorHeight, o.oPosZ,
                    nil
                )
            elseif distanceToPlayer < 2300 then
                spawn_non_sync_object(
                    id_bhvFlameLargeBurningOut,
                    E_MODEL_RED_FLAME,
                    o.oPosX, o.oFloorHeight, o.oPosZ,
                    nil
                )
            end
        elseif o.oForwardVel > 0 then
            o.oForwardVel = o.oForwardVel - 10
            cur_obj_move_using_fvel_and_gravity()
            if o.oForwardVel > 20 and cur_obj_lateral_dist_to_home() < 2300 then
                spawn_non_sync_object(
                    id_bhvFlameLargeBurningOut,
                    E_MODEL_RED_FLAME,
                    o.oPosX, o.oFloorHeight, o.oPosZ,
                    nil
                )
            end
        elseif o.oTimer > 30 then
            cur_obj_change_action(GRAND_BOSS_ACT_IDLE)
        end
    elseif o.oAction == GRAND_BOSS_ACT_SLAM then
        o.oFaceAnglePitch = 0
        o.oMoveAnglePitch = 0
        local homeDist = cur_obj_lateral_dist_to_home()
        o.oMoveAngleYaw = cur_obj_angle_to_home()
        if homeDist < 100 then
            o.oForwardVel = 0
        else
            o.oForwardVel = clamp(1, homeDist // 10, 50)
        end

        if o.oSubAction == 0 then
            if o.oHomeY + 2000 > o.oPosY then
                o.oVelY = (o.oHomeY - o.oPosY + 2010) // 10
            else
                o.oVelY = 0
            end

            if o.oTimer > 120 then
                o.oSubAction = 1
            end
        elseif o.oSubAction == 1 then
            o.oGravity = -4
        elseif o.oSubAction == 2 then
            if o.oVelY < 0 then
                o.oGravity = -8
            else
                o.oGravity = -4
                if o.oTimer % 5 == 0 then
                    cur_obj_play_sound_2(SOUND_MOVING_ALMOST_DROWNING)
                end
            end
        elseif o.oSubAction == 3 then
            o.oGravity = -4
            if o.oVelY < 0 then
                cur_obj_change_action(GRAND_BOSS_ACT_DIVE)
            end
        end
        cur_obj_move_using_fvel_and_gravity()
        o.oGravity = 0

        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1000

        if o.oFloorHeight >= o.oPosY and o.oVelY <= 0 then
            o.oVelY = 70
            o.oPosY = o.oFloorHeight
            cur_obj_play_sound_2(SOUND_GENERAL_BOWSER_BOMB_EXPLOSION)
            set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)

            spawn_non_sync_object(
                id_bhvBowserShockWave,
                E_MODEL_BOWSER_WAVE,
                o.oPosX, o.oPosY, o.oPosZ,
                nil
            )

            for i = 0, (MAX_PLAYERS - 1) do
                local m2 = gMarioStates[i]
                if is_player_active(m2) ~= 0 and o.oFloorHeight == m2.pos.y and m2.health > 0xFF then
                    set_mario_action(m2, ACT_THROWN_BACKWARD, 0)
                    if o.oSubAction == 2 then
                        m2.vel.y = 200
                    else
                        m2.vel.y = 80
                    end
                    m2.forwardVel = 0
                end
            end

            if o.oTimer > 210 then
                if o.oSubAction == 2 then
                    o.oVelY = 40
                    o.oSubAction = 3
                else
                    o.oVelY = 120
                    o.oSubAction = 2
                end
            end
        end
    elseif o.oAction == GRAND_BOSS_ACT_DYING then
        o.oIntangibleTimer = -1
        if o.oTimer == 1 and o.oSubAction == 0 then
            local m0 = gMarioStates[0]
            if m0.action & ACT_GROUP_CUTSCENE ~= 0 then -- fix issue when dying (camera is still messed up but at least there's no softlock)
                m0.health = 0x880
                m0.hurtCounter = 0
                set_mario_action(m0, ACT_IDLE, 0)
                reset_camera(m0.area.camera)
            end
            cutscene_object(CUTSCENE_STAR_SPAWN, o)
            o.activeFlags = o.activeFlags | ACTIVE_FLAG_INITIATED_TIME_STOP
            enable_time_stop_including_mario()
            fade_volume_scale(SEQ_PLAYER_LEVEL, 0, 30)
        end

        if o.oSubAction == 0 then
            if o.oTimer % 45 >= 35 then
                if o.oTimer >= 150 then
                    o.oSubAction = 1
                    o.oTimer = 0
                elseif o.oTimer % 45 == 35 then
                    cur_obj_play_sound_2(SOUND_GENERAL_BREAK_BOX)
                    set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)
                    spawn_triangle_break_particles(5, 0x79, 0.5, 50) -- MODEL_TRANSPARENT_STAR
                end
            end
        elseif o.oSubAction == 1 then
            o.oPosY = o.oPosY + 16 * coss(4000 * o.oTimer)
            if o.oTimer % 5 == 0 then
                cur_obj_play_sound_2(SOUND_ENV_STAR)
                set_camera_shake_from_hit(SHAKE_POS_LARGE)
            end
            if o.oTimer >= 50 then
                if o.oTimer % 25 == 0 then
                    cur_obj_play_sound_2(SOUND_GENERAL2_PYRAMID_TOP_EXPLOSION)
                    spawn_triangle_break_particles(5, 0x79, 0.5, 50) -- MODEL_TRANSPARENT_STAR
                end

                if o.oTimer >= 180 then
                    cur_obj_scale(0)
                else
                    cur_obj_scale(2 - (o.oTimer - 50) / 65)
                end
            end
            if o.oTimer > 180 then
                o.oSubAction = 2
                o.oTimer = 0
                cur_obj_play_sound_2(SOUND_GENERAL_GRAND_STAR)
                cur_obj_play_sound_2(SOUND_GENERAL2_PYRAMID_TOP_EXPLOSION)
                spawn_triangle_break_particles(10, 0x79, 1, 50) -- MODEL_TRANSPARENT_STAR
                set_camera_shake_from_hit(SHAKE_POS_LARGE)
                cur_obj_disable_rendering()
                cur_obj_scale(0)
            end
        elseif o.oSubAction == 2 then
            cur_obj_disable_rendering()
            cur_obj_scale(0)
            if o.oTimer > 90 then
                set_mario_action(gMarioStates[0], ACT_JUMBO_STAR_CUTSCENE, 0)
                disable_time_stop()
                gGlobalSyncTable.secretBoss = false
                obj_mark_for_deletion(o)
            end
        end
    end

    if o.oAction ~= GRAND_BOSS_ACT_HURT and o.oInteractStatus & INT_STATUS_TOUCHED_BOB_OMB ~= 0 then
        cur_obj_change_action(GRAND_BOSS_ACT_HURT)
        network_send_object(o, true)
    end
    o.oInteractStatus = 0
end

id_bhvGrandStarBoss = hook_behavior(nil, OBJ_LIST_GENACTOR, true, grand_star_boss_init, grand_star_boss_loop)

-- falling stars
function falling_star_init(o)
    o.oFlags = o.oFlags | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_HOLDABLE | OBJ_FLAG_COMPUTE_DIST_TO_MARIO

    local sFallingStarHitbox = get_temp_object_hitbox()
    sFallingStarHitbox.interactType = INTERACT_DAMAGE
    sFallingStarHitbox.downOffset = 0
    sFallingStarHitbox.damageOrCoinValue = 2
    sFallingStarHitbox.health = 3
    sFallingStarHitbox.numLootCoins = 0
    sFallingStarHitbox.radius = 80
    sFallingStarHitbox.height = 50
    sFallingStarHitbox.hurtboxRadius = 80
    sFallingStarHitbox.hurtboxHeight = 50
    obj_set_hitbox(o, sFallingStarHitbox)

    o.oIntangibleTimer = 0
    o.oFaceAngleYaw = 0x4000
    o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 40, -2, -0.5, 10, 10, 0
    network_init_object(o, true, {'oInteractType'})
end

function falling_star_loop(o)
    local flameModel = (o.oBehParams ~= 0 and E_MODEL_BLUE_FLAME) or E_MODEL_RED_FLAME
    if o.oHeldState == HELD_FREE then
        cur_obj_enable_rendering_and_become_tangible(o)
        if o.oAction == 0 then
            o.oForwardVel = 15
            o.oMoveAngleYaw = 0
            o.oFaceAngleRoll = o.oFaceAngleRoll + 0x1000
            o.oVelY = -30
            cur_obj_update_floor()
            cur_obj_move_standard(-58)
            if o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
                cur_obj_change_action(1)
            end

            for i = 0, 1 do
                spawn_non_sync_object(id_bhvKoopaShellFlame, flameModel, o.oPosX, o.oPosY, o.oPosZ, nil)
            end
        elseif o.oAction == 1 then
            if (o.oTimer <= 500 and (o.oBehParams ~= 0 or o.oTimer < 60)) and not (o.parentObj and o.parentObj.oAction == GRAND_BOSS_ACT_HURT) then
                for i = 0, 1 do
                    spawn_non_sync_object(id_bhvKoopaShellFlame, flameModel, o.oPosX, o.oPosY, o.oPosZ, nil)
                end
            else
                o.oInteractType = INTERACT_GRABBABLE
                if o.oTimer > 500 or (o.parentObj and o.parentObj.oAction == GRAND_BOSS_ACT_HURT) then
                    spawn_mist_particles()
                    if o.oPosX % 3 == 0 then
                        obj_spawn_yellow_coins(o, 1)
                    end
                    obj_mark_for_deletion(o)
                end
            end
        elseif o.oAction == 2 then
            o.oIntangibleTimer = 0
            cur_obj_update_floor_and_walls()
            cur_obj_move_standard(-58)
            if obj_attack_collided_from_other_object(o) == 1 or o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
                local m = nearest_mario_state_to_object(o)
                if m.playerIndex == 0 then
                    spawn_sync_object(id_bhvExplosion, E_MODEL_EXPLOSION, o.oPosX, o.oPosY, o.oPosZ, function(explosion)
                        explosion.oGraphYOffset = explosion.oGraphYOffset + 100
                    end)
                end
                obj_mark_for_deletion(o)
            end
        end
    elseif o.oHeldState == HELD_HELD then
        cur_obj_disable_rendering()
        o.oIntangibleTimer = -1
    elseif o.oHeldState == HELD_THROWN then
        o.oFaceAngleYaw = o.oMoveAngleYaw
        o.oFaceAngleRoll = 0
        o.oAction = 2
        o.oForwardVel = 25
        o.oVelY = 20
        o.oHeldState = 0
        cur_obj_enable_rendering_and_become_tangible(o)
    elseif o.oHeldState == HELD_DROPPED then
        spawn_mist_particles()
        obj_mark_for_deletion(o)
    end
    o.oInteractStatus = 0
end

id_bhvFallingStar = hook_behavior(nil, OBJ_LIST_DESTRUCTIVE, true, falling_star_init, falling_star_loop, "bhvFallingStar")
