-- name: Brutal Bosses
-- description: Makes all bosses, as well as various NPC-related missions, more difficult. And more... :)\n\nMod by EmilyEmmi\nBoth music ports by TheGael95\nKoopa The Nimble name from... Player\n+Support for Squishy's Character Select

-- makes this easy (by Isaac, used in mariohunt)
function hook_behavior_custom(id, override, init, loop)
  print("Hooked to id", tostring(id))
  if id == nil then return end
  local name = get_behavior_name_from_id(id) or ("bhvUnk" .. id)
  return hook_behavior(id,
    get_object_list_from_behavior(get_behavior_from_id(id)), -- Automatically get the correct object list
    override, init, loop,
    "bhvBossCustom" ..
    name:sub(4) -- Give the behavior a consistent behavior name (for example, bhvTreasureChestsJrb will become bhvBossCustomTreasureChestsJrb)
  )
end

-- Bowser stuff (this is really well programmed I swear)
function bhv_custom_bowser_loop(o)
  --print(o.oPrevAction, o.oAction, o.oSubAction, o.oTimer)

  -- fix bowser being active when he shouldn't (star road and ztar attack 2)
  if bhvStarRoadCustom004 or bhvPeachMessage then
    local np = gNetworkPlayers[0]
    if np.currAreaSyncValid and np.currLevelSyncValid and np.currLevelNum ~= LEVEL_BOWSER_3 then
      -- force bowser into do nothing action when not displaying cutscene
      if o.oAction ~= 5 and o.oAction ~= 6 and o.oAction ~= 20 then
        o.oAction = 4
        o.oSubAction = 11
        return
      end
    end
  end

  if (gGlobalSyncTable.secretBoss or _G.bossRushDarkStar) and o.oBehParams2ndByte == 2 then
    if o.oAction ~= 4 then
      o.oAction = 4
      o.oSubAction = 10
      o.oBowserUnkF8 = 2
      gMarioStates[0].area.camera.cutscene = 0
      play_cutscene(gMarioStates[0].area.camera)
    end
    o.oForwardVel = 0
    if o.oVelY > 0 then
      o.oVelY = -20
    end
  end

  if o.oAction == 0 then
    if o.oPrevAction == 5 then -- jump at start of fight
      cur_obj_change_action(17)
    elseif o.oPrevAction == 17 then
      local player = nearest_player_to_object(o)
      local dist = dist_between_objects(o, player) or 10000
      if dist > 1000 then
        cur_obj_change_action(7) -- run
        o.oMoveAngleYaw = obj_angle_to_object(o, player) or 0
      else
        o.oSubAction = 2
        cur_obj_change_action(9) -- quick fireball right after pound
      end
    elseif o.oPrevAction == 16 then
      cur_obj_change_action(13)     -- pound
      o.oSubAction = 1
    elseif o.oPrevAction == 19 then -- fire after stage pound
      cur_obj_change_action(15)
    end
  elseif o.oAction == 3 then -- breathe fire everywhere instead of dancing
    cur_obj_change_action(8)
  elseif o.oAction == 8 then -- make this action very fast
    if o.header.gfx.animInfo.animFrame < 23 then
      o.header.gfx.animInfo.animFrame = 23
    elseif o.header.gfx.animInfo.animFrame > 35 then
      cur_obj_change_action(0)
    end
  elseif o.oAction == 12 then -- teleport on hit
    o.oVelY = 40
    o.oForwardVel = 0
    cur_obj_change_action(16)
  elseif o.oAction == 19 or o.oAction == 13 then -- turn during pound to prevent grabbing immediately
    local player = nearest_player_to_object(o)
    obj_turn_toward_object(o, player, 16, 2048)
    local platform = (o.oAction == 19 and obj_get_nearest_object_with_behavior_id(o, id_bhvTiltingBowserLavaPlatform))
    if platform then
      platform.oAngleVelPitch = platform.oAngleVelPitch * 1.4
      platform.oAngleVelRoll = platform.oAngleVelRoll * 1.4
    end
  elseif o.oAction == 14 and o.oSubAction == 0 then -- do next action immediately
    o.oSubAction = 2
  elseif o.oAction == 15 then
    if o.header.gfx.animInfo.animFrame < 30 then -- speed up animation
      o.header.gfx.animInfo.animFrame = o.header.gfx.animInfo.animFrame + 5
    end
    if o.oBehParams2ndByte ~= 0 then -- spin whilst breathing fire (not in first fight)
      o.oFaceAngleYaw = o.oFaceAngleYaw + 0x600
      o.oMoveAngleYaw = o.oFaceAngleYaw
    end
  elseif o.oAction == 9 and o.oTimer < 5 then
    o.oBowserUnk108 = 2 + math.min(o.oBehParams2ndByte, 2) -- breathe fireballs based on fight number
    local player = nearest_player_to_object(o)
    o.oMoveAngleYaw = obj_angle_to_object(o, player) or 0
  elseif o.oAction == 7 then -- reduce run start and end lag
    if o.oSubAction == 0 then
      o.oSubAction = 1
    elseif o.oSubAction == 2 then
      cur_obj_change_action(0)
    end
    --djui_chat_message_create(tostring(o.oSubAction))
  elseif o.oAction == 10 and o.oBehParams2ndByte ~= 0 then -- teleport after reaching edge (not in first fight)
    cur_obj_change_action(16)
  elseif o.oAction == 16 and o.oSubAction ~= 0 then        -- teleport from above
    cur_obj_init_animation(9)
    local m = nearest_mario_state_to_object(o)
    o.oPosX = m.pos.x
    o.oPosZ = m.pos.z
    o.oPosY = m.pos.y + 1000
    if m.playerIndex == 0 then
      cur_obj_play_sound_2(SOUND_MOVING_ALMOST_DROWNING)
    end
    if m.action == ACT_RELEASING_BOWSER and o.oBowserUnkF8 < 30 then -- give you time to move
      o.oBowserUnkF8 = 30
    end
  elseif o.oAction == 1 and o.oSubAction == 1 then
    if o.oBehParams2ndByte == 0 then -- if you don't get him, he does the fire spin (jumps in first fight)
      cur_obj_change_action(17)
      local player = nearest_player_to_object(o)
      o.oMoveAngleYaw = obj_angle_to_object(o, player) or 0
    else
      cur_obj_change_action(15)
    end
  elseif o.oAction == 2 then
    if o.oSubAction == 3 or (o.oSubAction == 2 and o.oBowserDistToCentre < 2500 and math.abs(o.oHomeY - o.oFloorHeight) < 100) then
      local player = nearest_player_to_object(o) -- turn to prevent grabbing
      obj_turn_toward_object(o, player, 16, 2048)
    end
    
    if o.oPrevAction == 17 then -- fix softlock
      cur_obj_change_action(16)
      o.oPosY = o.oPosY + 5000
    end
  elseif IS_VANILLA and o.oAction == 4 and o.oSubAction == 10 and o.oBowserUnkF8 == 1 then
    if gGlobalSyncTable.darkStarFlags ~= 7 and not _G.bossRushDarkStar then
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.Bowser3Defeated120StarsDialog, 1, 4, 30, 200,
        DIALOG_BOWSER_120)
    else
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.Bowser3Defeated120StarsDialog, 1, 4, 30, 200,
        DIALOG_BOWSER_SECRET)
      gGlobalSyncTable.secretBoss = true
    end
  end
end

function bhv_custom_bowser_init(o)
  o.oHealth = 3 -- always 3 hits
end

-- rip bowser platforms
function delete(o)
  o.activeFlags = ACTIVE_FLAG_DEACTIVATED
end

hook_behavior_custom(id_bhvBowser, false, bhv_custom_bowser_init, bhv_custom_bowser_loop)
--hook_behavior_custom(bhvStarRoadCustom004, false, bhv_custom_bowser_init, bhv_custom_bowser_loop) -- TODO: Star road
hook_behavior_custom(id_bhvFallingBowserPlatform, true, delete)

-- King Bob-Omb
--- @param o Object
function bhv_custom_king_bobomb_loop(o)
  o.hitboxHeight = 200 -- hitting with bombs is easier
  o.oBounciness = 0 -- prevent falling off after jump throw
  --print(o.oPrevAction, o.oAction, o.oSubAction, o.oTimer, o.oKingBobombUnk88)

  if o.oAction == 2 then
    local angleToLocalMario = obj_angle_to_point(o, gMarioStates[0].pos.x, gMarioStates[0].pos.z)
    if abs_angle_diff(o.oFaceAngleYaw, angleToLocalMario) > 0x2AAA then -- prevent grab
      o.oInteractType = 8
    else
      o.oInteractType = 2
    end
    o.oKingBobombUnkF8 = 0 -- prevents falling
    o.oKingBobombUnk88 = 3 -- only used when grabbing mario, so it can be used here too
    o.oForwardVel = o.oForwardVel * 3
    local player = nearest_player_to_object(o)
    local angleToPlayer = obj_angle_to_object(o, player) or 0
    cur_obj_rotate_yaw_toward(angleToPlayer, 0x500)
    if o.oTimer > 120 and o.oKingBobombUnk100 == 2 then
      cur_obj_init_animation_and_anim_frame(9, 16) -- start at throw part
      cur_obj_change_action(9)                     -- throw bomb action
    end
    local bomb = king_bobomb_test_bomb_collision(o)
    if bomb then
      if is_nearest_mario_state_to_object(gMarioStates[0], o) ~= 0 then
        local explosion = spawn_sync_object(
          id_bhvExplosion,
          E_MODEL_EXPLOSION,
          bomb.oPosX, bomb.oPosY, bomb.oPosZ,
          nil
        )

        if explosion then
          explosion.oGraphYOffset = explosion.oGraphYOffset + 100
        end
      end

      bomb.activeFlags = ACTIVE_FLAG_DEACTIVATED

      cur_obj_init_animation_with_sound(2)
      cur_obj_change_action(10)
      o.oTimer = 0
      network_send_object(o, true)
    end
  elseif o.oAction == 6 and o.oSubAction == 2 then -- switch to charge action
    cur_obj_init_animation_with_sound(11)
    cur_obj_become_tangible()
    cur_obj_change_action(12)
  elseif o.oAction == 12 then
    local player = nearest_player_to_object(o)
    local angleToPlayer = obj_angle_to_object(o, player) or 0
    local dist = lateral_dist_between_objects(o, player) or 10000
    if o.oTimer < 70 then -- charge towards player after hit
      cur_obj_rotate_yaw_toward(angleToPlayer, 0x800)
      o.oForwardVel = -5
    else
      o.hitboxHeight = 600
      o.oKingBobombUnk88 = 1
      cur_obj_rotate_yaw_toward(angleToPlayer, 0x300)
      obj_init_animation_with_accel_and_sound(o, 1, 5)

      o.oInteractType = 2

      if o.oSubAction == 1 then
        o.oForwardVel = o.oForwardVel - 10
      else
        o.oForwardVel = o.oForwardVel + 20
      end
      --cur_obj_init_animation_and_anim_frame(1,o.oTimer - 61)
      if o.oMoveFlags & OBJ_MOVE_HIT_EDGE ~= 0 then
        o.oMoveAngleYaw = o.oMoveAngleYaw + 0x8000
        o.oForwardVel = 3
        if o.usingObj then
          o.oKingBobombUnk88 = 1
          cur_obj_change_action(3)
        else
          cur_obj_change_action(2)
        end
      elseif o.oForwardVel < 3 then
        o.oForwardVel = 3
        if o.usingObj then
          o.oKingBobombUnk88 = 1
          cur_obj_change_action(3)
        else
          cur_obj_change_action(2)
        end
      elseif o.oSubAction == 0 and (o.usingObj or dist < 100) then
        o.oSubAction = 1
      end
    end
  elseif o.oAction == 3 and o.oSubAction ~= 0 then -- jump when throwing player
    if o.usingObj == gMarioStates[0].marioObj and gMarioStates[0].action == ACT_GRABBED then
      gMarioStates[0].pos.x, gMarioStates[0].pos.y, gMarioStates[0].pos.z = o.oPosX, o.oPosY, o.oPosZ -- so the camera follows
    end

    if o.oVelY > 0 then
      local prevYaw = o.oMoveAngleYaw
      o.oMoveAngleYaw = limit_angle(o.oMoveAngleYaw + 0x1000)
      if o.oMoveAngleYaw < 0 and prevYaw > 0 then
        play_sound(SOUND_OBJ_BOWSER_SPINNING, o.header.gfx.cameraToObject)
      end
    end
    
    if o.oSubAction == 2 then
      if o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND == 0 then
        cur_obj_extend_animation_if_at_end()
      end
    elseif o.oTimer == 24 then
      o.oVelY = 100
    elseif o.oTimer > 30 then
      o.oSubAction = 2
    end
  elseif o.oAction == 9 then -- new throw bomb action
    o.oInteractType = 8
    o.oForwardVel = 0
    if o.oTimer == 12 and is_nearest_mario_state_to_object(gMarioStates[0], o) ~= 0 then
      spawn_sync_object(
        id_bhvBobomb,
        E_MODEL_BLACK_BOBOMB,
        o.oPosX + 250 * sins(o.oMoveAngleYaw), o.oPosY, o.oPosZ + 250 * coss(o.oMoveAngleYaw),
        function(bomb)
          bomb.oBobombFuseLit = 1
          bomb.oFaceAngleYaw = o.oFaceAngleYaw
          bomb.oMoveAngleYaw = o.oMoveAngleYaw
          bomb.oForwardVel = 50
          bomb.oAction = 2
          bomb.oVelY = 20
          bomb.oBehParams = 0x100 -- disable coin
        end)
    elseif o.oTimer > 30 then
      if o.oKingBobombUnk88 <= o.oHealth then
        cur_obj_change_action(2)
      else
        cur_obj_init_animation_and_anim_frame(9, 16) -- start at throw part
        cur_obj_change_action(9)                     -- throw again for each hit point
        o.oKingBobombUnk88 = o.oKingBobombUnk88 - 1
      end
    end
    local bomb = king_bobomb_test_bomb_collision(o)
    if bomb then
      local m = nearest_mario_state_to_object(o)
      if m.playerIndex == 0 then
        local explosion = spawn_sync_object(
          id_bhvExplosion,
          E_MODEL_EXPLOSION,
          bomb.oPosX, bomb.oPosY, bomb.oPosZ,
          nil
        )

        if explosion then
          explosion.oGraphYOffset = explosion.oGraphYOffset + 100
        end
      end

      bomb.activeFlags = ACTIVE_FLAG_DEACTIVATED
      
      cur_obj_init_animation_with_sound(2)
      cur_obj_change_action(10)
      o.oTimer = 0
      network_send_object(o, true)
    end
  elseif o.oAction == 10 then -- new stunned action
    cur_obj_init_animation_with_sound(2)
    o.oForwardVel = 0
    o.oInteractType = 2
    o.oForwardVel = 0
    if o.usingObj then
      o.oKingBobombUnk88 = 1
      cur_obj_change_action(3)
    elseif o.oTimer > 90 then -- 3 seconds
      cur_obj_change_action(6)
    end
  elseif o.oAction == 8 and o.oSubAction == 0 then -- switch to new end action
    local m = nearest_mario_state_to_object(o)
    if m then
      local dist = dist_between_object_and_point(o, m.pos.x, m.pos.y, m.pos.z)
      o.oForwardVel = dist / 20
      o.oMoveAngleYaw = obj_angle_to_point(o, m.pos.x, m.pos.z) or 0
      o.oVelY = 40
      cur_obj_change_action(11)
      network_send_object(o, true)
    end
  elseif o.oAction == 11 and o.oFloorHeight >= o.oPosY then -- new end action
    cur_obj_change_action(8)
    o.oSubAction = 1
    if is_nearest_mario_state_to_object(gMarioStates[0], o) ~= 0 then
      spawn_sync_object(
        id_bhvExplosion,
        E_MODEL_EXPLOSION,
        o.oPosX, o.oPosY, o.oPosZ,
        nil)
      network_send_object(o, true)
    end
  end
end

function king_bobomb_test_bomb_collision(o)
  --if (o.oInteractStatus & INT_STATUS_TOUCHED_BOB_OMB) ~= 0 then return true end
  if o.numCollidedObjs > 0 then
    for i = 0, (o.numCollidedObjs - 1) do
      local bomb = obj_get_collided_object(o, i)
      if bomb and get_id_from_behavior(bomb.behavior) == id_bhvBobomb and bomb.oAction == 1 then
        return bomb
      end
      i = i + 1
    end
  end
end

hook_behavior_custom(id_bhvKingBobomb, false, nil, bhv_custom_king_bobomb_loop)

-- Whomp King
function bhv_custom_king_whomp_loop(o)
  cur_obj_scale(4)            -- double size
  o.oCollisionDistance = 2000 -- double size
  o.hurtboxRadius = 10000
  local player = nearest_player_to_object(o)
  local angleToPlayer = obj_angle_to_object(o, player) or 0
  if o.oAction == 3 and o.unused1 == 1 then
    cur_obj_change_action(10)
  elseif o.oAction == 4 then
    local m = nearest_possible_mario_state_to_object(o)
    if m.action == ACT_GROUND_POUND then
      cur_obj_change_action(4)
    end
  elseif o.unused1 == 1 and o.oAction == 2 and o.oTimer > 120 then
    cur_obj_change_action(10) -- alt attack
  elseif o.oAction == 2 and o.oPrevAction == 2 and abs_angle_diff(o.oMoveAngleYaw, angleToPlayer) < 0x200 and o.oTimer > 60 then
    cur_obj_change_action(3)
  elseif o.oAction == 6 then
    o.unused1 = 1
    if o.oPrevAction == 5 then -- squish doesn't work for some reason, so I wrote my own
      set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)
      spawn_non_sync_object(
        id_bhvBowserShockWave,
        E_MODEL_BOWSER_WAVE,
        o.oPosX + 600 * sins(o.oMoveAngleYaw), o.oFloorHeight, o.oPosZ + 600 * coss(o.oMoveAngleYaw),
        nil
      )

      for i = 0, (MAX_PLAYERS - 1) do
        local m = gMarioStates[i]
        if m.health > 0xFF then
          if m.ceil and m.ceil.object == o and m.squishTimer > 0 then
            play_character_sound(m, CHAR_SOUND_ATTACKED)
            m.hurtCounter = 0x24 -- instant death
          elseif o.oFloorHeight == m.pos.y and m.invincTimer == 0 and (m.action & ACT_FLAG_INVULNERABLE) == 0 then
            o.oDamageOrCoinValue = 2
            o.oInteractionSubtype = o.oInteractionSubtype & ~INT_SUBTYPE_DELAY_INVINCIBILITY
            take_damage_and_knock_back(m, o)
            o.oDamageOrCoinValue = 8
          end
        end
      end
    elseif o.oTimer > 60 then
      o.oSubAction = 10       -- come back up
    end
  elseif o.oAction == 10 then -- custom action
    o.oForwardVel = 0
    cur_obj_init_animation_and_anim_frame(1, o.oTimer)
    if o.oTimer == 10 then
      o.oVelY = 50
    elseif o.oTimer > 30 and o.oPosY - o.oFloorHeight < 1 then
      set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)
      cur_obj_play_sound_2(SOUND_OBJ_WHOMP_LOWPRIO)
      for i = 0, (MAX_PLAYERS - 1) do
        local m = gMarioStates[i]
        if o.oPosY == m.pos.y and m.pos.y - m.floorHeight < 1 and m.health > 0xFF then
          hurt_and_set_mario_action(m, ACT_FORWARD_GROUND_KB, 0, 0x8) -- 2 damage
        end
      end

      -- mini whomps
      if is_nearest_mario_state_to_object(gMarioStates[0], o) ~= 0 then
        spawn_sync_object(
          id_bhvSmallWhomp,
          E_MODEL_WHOMP,
          o.oPosX + 500 * sins(o.oMoveAngleYaw), o.oFloorHeight, o.oPosZ + 500 * coss(o.oMoveAngleYaw),
          function(whomp)
            whomp.oBehParams = 0x1000 -- set as mini whomp
            whomp.oVelY = 50          -- jump up
            whomp.parentObj = o
          end)
      end

      o.unused1 = 0
      cur_obj_change_action(2)
    end
  end
end

function bhv_custom_king_whomp_init(o)
  o.unused1 = 0
end

function bhv_custom_whomp_loop(o)
  if o.oBehParams == 0x1000 then
    if o.oAction == 6 or o.parentObj.oAction == 8 then
      cur_obj_change_action(8) -- die
    elseif o.oAction == 0 or o.oAction == 2 then
      cur_obj_change_action(1)
    elseif o.oAction == 1 then
      o.oHomeX = o.oPosX
      o.oHomeZ = o.oPosZ
      local player = nearest_player_to_object(o)
      local angleToPlayer = obj_angle_to_object(o, player) or 0
      cur_obj_rotate_yaw_toward(angleToPlayer, 0x200)
    elseif o.oAction == 4 then
      o.oForwardVel = 10       -- jump forward a bit to hit Mario
    end
    cur_obj_scale(0.5)         -- half size
    o.oCollisionDistance = 250 -- half size
    o.hurtboxRadius = 1250
  end
end

function bhv_custom_whomp_init(o)
  if o.oBehParams == 0x1000 then
    o.collisionData          = nil
    o.oInteractType          = INTERACT_DAMAGE
    o.oNumLootCoins          = 0
    local hitbox             = get_temp_object_hitbox()
    hitbox.interactType      = INTERACT_DAMAGE
    hitbox.downOffset        = 0
    hitbox.damageOrCoinValue = 2
    hitbox.health            = 1
    hitbox.numLootCoins      = 0
    hitbox.radius            = 40
    hitbox.height            = 200
    hitbox.hurtboxRadius     = 40
    hitbox.hurtboxHeight     = 200
    obj_set_hitbox(o, hitbox)
  end
end

hook_behavior_custom(id_bhvWhompKingBoss, false, bhv_custom_king_whomp_init, bhv_custom_king_whomp_loop)
hook_behavior_custom(id_bhvSmallWhomp, false, bhv_custom_whomp_init, bhv_custom_whomp_loop)

-- fix for shockwave (no longer damages when not on same floor as shockwave)
local D_8032F420 = { 1.9, 2.4, 4.0, 4.8 }
function custom_shock_wave_init(o)
  o.oFlags = o.oFlags | OBJ_FLAG_ACTIVE_FROM_AFAR | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
  o.oOpacity = 255
  o.oFaceAngleYaw = 0
  o.oFaceAnglePitch = 0
  o.oFaceAngleRoll = 0
end

function custom_shock_wave_loop(o)
  local sp2C, sp28, sp24, sp20;
  local sp1E = 70;
  o.oBowserShockWaveUnkF4 = o.oTimer * 10;
  cur_obj_scale(o.oBowserShockWaveUnkF4);
  if (o.oTimer % 3) ~= 0 then -- this used to use global timer but can't do that obviously
    o.oOpacity = o.oOpacity - 1
  end
  if (o.oTimer > sp1E) then
    o.oOpacity = o.oOpacity - 5
  end
  if (o.oOpacity <= 0) then
    obj_mark_for_deletion(o)
  end
  if (o.oTimer < sp1E and mario_is_in_air_action(gMarioStates[0]) == 0 and o.oPosY == gMarioStates[0].floorHeight) then
    sp2C = o.oBowserShockWaveUnkF4 * D_8032F420[1];
    sp28 = o.oBowserShockWaveUnkF4 * D_8032F420[2];
    sp24 = o.oBowserShockWaveUnkF4 * D_8032F420[3];
    sp20 = o.oBowserShockWaveUnkF4 * D_8032F420[4];
    if ((sp2C < o.oDistanceToMario and o.oDistanceToMario < sp28)
          or (sp24 < o.oDistanceToMario and o.oDistanceToMario < sp20)) then
      gMarioStates[0].marioObj.oInteractStatus = gMarioStates[0].marioObj.oInteractStatus | INT_STATUS_HIT_BY_SHOCKWAVE
    end
  end
end

hook_behavior_custom(id_bhvBowserShockWave, true, custom_shock_wave_init, custom_shock_wave_loop)

-- big boo
--- @param o Object
function custom_big_boo_init(o)
  -- boos randomly have their speed and turning speed reduced; this disables that
  o.oBooNegatedAggressiveness = 0
  o.oBooTurningSpeed = 128
end

--- @param o Object
function custom_big_boo_loop(o)
  local difficulty = o.oHealth - o.oBehParams2ndByte + 2
  if difficulty < 0 then
    difficulty = 1
  end

  local player = nearest_player_to_object(o)
  if (o.oBooTargetOpacity == 40 or o.oAction == 3 or (o.oAction == 2 and o.oBehParams2ndByte ~= 1)) and o.oHealth ~= 0 and player and dist_between_objects(o, player) < 1500 then
    o.oBooTargetOpacity = 0 -- fully disappear when:
    -- looked at
    -- hit
    -- jumped on (unless this is the merry go round boo because there's a stupid bug)
  elseif o.oOpacity == 255 and (o.oTimer > (difficulty * 40) or o.oBooNegatedAggressiveness == 1) and o.oAction == 1 then -- spit fire at intevals determined by health and type
    if player and dist_between_objects(o, player) < 1500 then
      local mouthDistanceH = o.oBooBaseScale * 40
      local mouthDistanceV = o.oBooBaseScale * 20
      cur_obj_play_sound_2(SOUND_OBJ_FLAME_BLOWN)
      obj_spit_fire(0, mouthDistanceV, mouthDistanceH, 5, 0x91, o.oForwardVel + 15, 10, 0) -- 0x91 is blue flame model

      if o.oBooNegatedAggressiveness > (4 - o.oHealth) then
        o.oBooNegatedAggressiveness = 0
        o.oTimer = 0
      elseif o.oBooNegatedAggressiveness ~= 0 then
        o.oBooNegatedAggressiveness = o.oBooNegatedAggressiveness + 1 -- used as counters
        o.oBooTargetOpacity = 0
      else
        o.oTimer = 0
      end
    else
      o.oTimer = 0
    end
  end

  if o.oOpacity == 0 and o.oAction == 1 then -- teleport behind mario when fully transparent
    local m = nearest_mario_state_to_object(o)
    o.oBooTargetOpacity = 255
    if m and (o.activeFlags & ACTIVE_FLAG_IN_DIFFERENT_ROOM) == 0 then
      o.oFaceAngleYaw = o.oFaceAngleYaw + 0x3000
      o.oMoveAngleYaw = o.oFaceAngleYaw

      local origX = o.oPosX
      local origZ = o.oPosZ
      o.oPosX = m.pos.x
      o.oPosZ = m.pos.z
      o.oVelX = (-sins(o.oFaceAngleYaw) * (o.hitboxRadius + 300)) / 4
      o.oVelZ = (-coss(o.oFaceAngleYaw) * (o.hitboxRadius + 300)) / 4
      local try = 0
      local try2 = 0
      while try < 4 and try2 < 4 do -- 4 steps, similar to mario
        if cur_obj_move_xz(200, 0) == 0 then
          o.oFaceAngleYaw = o.oFaceAngleYaw + 0x3000
          o.oMoveAngleYaw = o.oFaceAngleYaw
          o.oPosX = m.pos.x
          o.oPosZ = m.pos.z
          o.oVelX = (-sins(o.oFaceAngleYaw) * (o.hitboxRadius + 300)) / 4
          o.oVelZ = (-coss(o.oFaceAngleYaw) * (o.hitboxRadius + 300)) / 4

          try = 0
          try2 = try2 + 1
        else
          try = try + 1
        end
      end
      if try2 >= 4 then
        o.oPosX, o.oPosZ = origX, origZ
      end

      if o.oBooNegatedAggressiveness == 0 then
        o.oBooNegatedAggressiveness = 1
      end
    end
  end
end

hook_behavior_custom(id_bhvMerryGoRoundBigBoo, false, custom_big_boo_init, custom_big_boo_loop)
hook_behavior_custom(id_bhvBalconyBigBoo, false, custom_big_boo_init, custom_big_boo_loop)
hook_behavior_custom(id_bhvGhostHuntBigBoo, false, custom_big_boo_init, custom_big_boo_loop)
hook_behavior_custom(id_bhvBooWithCage, false, custom_big_boo_init, custom_big_boo_loop)

-- bullies
local BULLY_ACT_RECOVER = 8 -- custom
local BULLY_ACT_CHARGE = 9
local BULLY_ACT_JUMP = 10
--- @param o Object
function custom_bully_init(o)
  o.oHealth = 3
end

--- @param o Object
function custom_bully_loop(o)
  if o.oAction == BULLY_ACT_CHASE_MARIO then -- always attack player
    if o.oPrevAction == BULLY_ACT_KNOCKBACK then
      cur_obj_change_action(BULLY_ACT_CHARGE)
      return
    end
    o.oTimer = 11
    o.oFaceAngleYaw = o.oMoveAngleYaw
    local player = nearest_player_to_object(o)
    if player then
      obj_turn_toward_object(o, player, 16, 2048)
    end
  elseif o.oAction == BULLY_ACT_KNOCKBACK then -- recover at edge first time (also reduces knockback)
    if o.oPrevAction == BULLY_ACT_CHARGE then
      local index = network_local_index_from_global(o.oBullyLastNetworkPlayerIndex)
      if index then
        local m = gMarioStates[index]
        if m.action & ACT_FLAG_INVULNERABLE ~= 0 then
          o.oForwardVel = 0
          local angleToPlayer = obj_angle_to_object(o, m.marioObj)
          m.faceAngle.y = angleToPlayer
          set_mario_action(m, ACT_THROWN_FORWARD, 0)
          m.vel.y = 40
          m.forwardVel = 50
          --m.faceAngle.y = o.oBullyMarioCollisionAngle
        end
      end
    end

    if o.oTimer > 5 and o.oVelY < 0 then
      o.oPosX = o.oBullyPrevX
      o.oPosY = o.oBullyPrevY
      o.oPosZ = o.oBullyPrevZ
      o.oForwardVel = 0
      o.oTimer = 0
    end
    if o.oBullyKBTimerAndMinionKOCounter == 10 then
      if o.oForwardVel > 0 and math.floor(o.oPosX) % 3 == 0 then -- random-ish
        o.oVelY = 100
        o.oPosX = o.oPosX + 1
        o.oForwardVel = 0
        cur_obj_play_sound_2(SOUND_GENERAL_ENEMY_ALERT1)
        cur_obj_change_action(BULLY_ACT_JUMP)
      end
      o.oBullyKBTimerAndMinionKOCounter = 17                      -- recover faster
    end
  elseif o.oAction == BULLY_ACT_LAVA_DEATH and o.oHealth > 1 then -- recover
    cur_obj_play_sound_2(SOUND_OBJ_BULLY_EXPLODE_2)
    o.oHealth = o.oHealth - 1
    o.oPosY = o.oPosY + 200

    local ydist = cur_obj_abs_y_dist_to_home() + 300
    local frames = 30
    o.oVelY = ydist / frames + 4 * frames

    o.oBullyKBTimerAndMinionKOCounter = 0
    o.oForwardVel = 0
    cur_obj_become_intangible()
    cur_obj_change_action(BULLY_ACT_RECOVER)
  elseif o.oAction == BULLY_ACT_RECOVER then
    local collisionFlags = object_step()
    if o.oVelY <= 10 and o.oForwardVel == 0 then
      o.oAngleToHome = cur_obj_angle_to_home()
      o.oMoveAngleYaw = o.oAngleToHome
      o.oFaceAngleYaw = o.oMoveAngleYaw
      o.oForwardVel = 30
    end
    if ((collisionFlags & 0x9) == 0x9) then -- bits 0 and 3
      cur_obj_init_animation(1)
      cur_obj_change_action(BULLY_ACT_CHARGE)
      o.oTimer = 0
    end
    if (collisionFlags == 1) then
      cur_obj_become_tangible()
      cur_obj_play_sound_2(SOUND_OBJ_THWOMP);
      set_camera_shake_from_point(SHAKE_POS_SMALL, o.oPosX, o.oPosY, o.oPosZ)
      spawn_mist_particles()
    end
  elseif o.oAction == BULLY_ACT_CHARGE then
    local collisionFlags = object_step()
    if (collisionFlags & 0x8) == 0 then -- bully backup check
      o.oPosX = o.oBullyPrevX;
      o.oPosZ = o.oBullyPrevZ;
      o.oAction = BULLY_ACT_BACK_UP;
    elseif o.oTimer < 11 then
      o.oForwardVel = 5
      local player = nearest_player_to_object(o)
      if player then
        obj_turn_toward_object(o, player, 16, 2048)
      end
      o.oFaceAngleYaw = o.oMoveAngleYaw
    elseif o.oTimer > 35 then
      o.oTimer = 0
      cur_obj_change_action(BULLY_ACT_CHASE_MARIO)
    else
      o.oForwardVel = 70
    end
  elseif o.oAction == BULLY_ACT_JUMP then
    local collisionFlags = object_step()
    if (collisionFlags == 1) then
      cur_obj_init_animation(1)
      o.oVelY = 0

      spawn_non_sync_object(
        id_bhvBowserShockWave,
        E_MODEL_BOWSER_WAVE,
        o.oPosX, o.oPosY, o.oPosZ,
        nil
      )

      cur_obj_change_action(BULLY_ACT_CHARGE)
      cur_obj_play_sound_2(SOUND_OBJ_THWOMP);
      set_camera_shake_from_point(SHAKE_POS_SMALL, o.oPosX, o.oPosY, o.oPosZ)
      spawn_mist_particles()
    end
  end
end

hook_behavior_custom(id_bhvBigBully, false, custom_bully_init, custom_bully_loop)
hook_behavior_custom(id_bhvBigBullyWithMinions, false, custom_bully_init, custom_bully_loop)
hook_behavior_custom(id_bhvBigChillBully, false, custom_bully_init, custom_bully_loop)
--hook_behavior_custom(bhvSMSRBigBully, false, custom_bully_init, custom_bully_loop)
--hook_behavior_custom(bhvSMSRBigChillBully, false, custom_bully_init, custom_bully_loop)
--hook_behavior_custom(bhvStarRoadCustom003, false, custom_bully_init, custom_bully_loop)

-- Mr. I
--- @param o Object
function custom_eye_init(o)
  if o.oBehParams2ndByte == 0 then return end

  o.oHealth = 3
  --o.oPosX = 200
  --o.oPosY = -100
  --o.oPosZ = 5000
  --o.oRoom = 0
end

--- @param o Object
function custom_eye_loop(o)
  if o.oBehParams2ndByte == 0 then return end

  local m = nearest_mario_state_to_object(o)

  o.oMrIUnk104 = 0
  if m and m.playerIndex == 0 and (o.oAction == 1 or o.oAction == 2) then
    if (m.flags & MARIO_VANISH_CAP) ~= 0 then
      cur_obj_change_action(1)                                  -- don't see
    elseif m and m.invincTimer ~= 0 then
      o.oMrIUnkFC = 0                                           -- reset eye progress
    end
    if o.oTimer > 15 and (o.oAction == 2 or o.oTimer > 30) then -- shoot a lot
      spawn_mr_i_particle(o)
      if o.oAction == 1 then                                    -- 3 particles!?
        spawn_mr_i_particle(o, o.oMoveAngleYaw - 0x1000)
        spawn_mr_i_particle(o, o.oMoveAngleYaw + 0x1000)
      end
      o.oTimer = 1
    end
  elseif o.oAction == 3 and o.oPrevAction == 3 then
    if m and m.playerIndex == 0 and o.oTimer <= 50 and o.oTimer % 5 == 0 then
      spawn_mr_i_particle(o)
    end
    if o.oTimer > 60 and o.oHealth > 1 then -- survive hit
      cur_obj_play_sound_2(SOUND_OBJ2_WHOMP_SOUND_SHORT)
      o.oHealth = o.oHealth - 1
      o.oMrIUnkFC = 0
      cur_obj_change_action(2)
    elseif o.oTimer == 104 and m and m.playerIndex == 0 then
      for i = 0, 7 do
        spawn_mr_i_particle(o, (0x2000 * i)) -- final attack
      end
    end
  end
end

-- recreation
function spawn_mr_i_particle(o, angle_)
  local angle = angle_ or o.oMoveAngleYaw
  local sp18 = o.header.gfx.scale.y
  local particle = spawn_sync_object(
    id_bhvMrIParticle,
    E_MODEL_PURPLE_MARBLE,
    o.oPosX, o.oPosY, o.oPosZ,
    nil
  )
  if particle then
    particle.oPosY = particle.oPosY + 50 * sp18;
    particle.oPosX = particle.oPosX + sins(angle) * 90 * sp18;
    particle.oPosZ = particle.oPosZ + coss(angle) * 90 * sp18;
    particle.oMoveAngleYaw = angle
  end

  cur_obj_play_sound_2(SOUND_OBJ_MRI_SHOOT)
end

hook_behavior_custom(id_bhvMrI, false, custom_eye_init, custom_eye_loop)

-- wiggler
local WIGGLER_ACT_CUSTOM_SHRINK = WIGGLER_ACT_FALL_THROUGH_FLOOR + 1
local WIGGLER_ACT_CUSTOM_GROW = WIGGLER_ACT_FALL_THROUGH_FLOOR + 2
local E_MODEL_WIGGLER_HEAD_MAD = smlua_model_util_get_id("wiggler_head_mad_geo")
local E_MODEL_WIGGLER_BODY_MAD = smlua_model_util_get_id("wiggler_body_mad_geo")
--- @param o Object
function custom_wiggler_loop(o)
  if (o.unused1 ~= 1 and o.unused1 ~= 2) and o.oHealth == 2 and o.oAction ~= WIGGLER_ACT_CUSTOM_SHRINK and o.header.gfx.scale.x ~= 1 then
    obj_set_model_extended(o, E_MODEL_WIGGLER_HEAD_MAD)
    o.unused1 = 1
  elseif o.unused1 == 1 and o.oAction == WIGGLER_ACT_JUMPED_ON then
    obj_set_model_extended(o, E_MODEL_WIGGLER_HEAD)
    o.unused1 = 2
  end

  if o.oAction == WIGGLER_ACT_WALK and o.oHealth ~= 1 then
    if o.header.gfx.scale.x == 1 and o.oVelY == 0 then
      cur_obj_play_sound_2(SOUND_OBJ_WIGGLER_JUMP)
      o.oForwardVel = 30
      o.oVelY = 50
      if is_nearest_mario_state_to_object(gMarioStates[0], o) then
        spawn_mr_i_particle(o)                                                             -- getting a lot of use out of this
      end
    elseif o.oHealth ~= 4 and o.oTimer % ((o.oHealth - 1) * 30) == 0 and o.oVelY == 0 then -- jump occasionally
      cur_obj_play_sound_2(SOUND_OBJ_WIGGLER_JUMP)
      o.oForwardVel = 40
      o.oVelY = 50

      if is_nearest_mario_state_to_object(gMarioStates[0], o) then -- final part
        spawn_mr_i_particle(o, o.oMoveAngleYaw - 0x1000)
        spawn_mr_i_particle(o)
        spawn_mr_i_particle(o, o.oMoveAngleYaw + 0x1000)
      end
    end

    if o.oHealth == 2 and o.header.gfx.scale.x ~= 1 and o.oForwardVel > 20 and (o.oTimer % 10 == 0) then -- FIRE
      spawn_non_sync_object(
        id_bhvFlameLargeBurningOut,
        E_MODEL_RED_FLAME,
        o.oPosX - sins(o.oMoveAngleYaw) * 100, o.oFloorHeight, o.oPosZ - coss(o.oMoveAngleYaw) * 100,
        nil
      )
    end
  elseif o.oAction == WIGGLER_ACT_KNOCKBACK and o.header.gfx.scale.x == 1 and o.oForwardVel <= 1 then -- grow back when hit in small form
    cur_obj_play_sound_2(SOUND_MENU_EXIT_PIPE)
    o.oAction = WIGGLER_ACT_CUSTOM_GROW
  elseif o.oAction == WIGGLER_ACT_JUMPED_ON and o.oTimer >= 29 and o.oHealth > 2 then
    --djui_popup_create(tostring(o.oHealth), 1)
    o.oHealth = o.oHealth - 1
    o.oAction = WIGGLER_ACT_CUSTOM_SHRINK
  elseif o.oAction == WIGGLER_ACT_CUSTOM_SHRINK then
    wiggler_custom_act_shrink(o)
  elseif o.oAction == WIGGLER_ACT_CUSTOM_GROW then
    wiggler_custom_act_grow(o)
  end
end

function custom_wiggler_body_loop(o) -- only does model switch
  if not o.parentObj then return end
  if (o.unused1 ~= 1 and o.unused1 ~= 2) and o.parentObj.unused1 == 1 then
    obj_set_model_extended(o, E_MODEL_WIGGLER_BODY_MAD)
    o.unused1 = 1
  elseif o.unused1 == 1 and o.parentObj.unused1 == 2 then
    obj_set_model_extended(o, E_MODEL_WIGGLER_BODY)
    o.unused1 = 2
  end
end

-- like standard shrink but doesn't spawn star
function wiggler_custom_act_shrink(o)
  if (o.oTimer >= 20) then
    if (o.oTimer == 20) then
      cur_obj_play_sound_2(SOUND_OBJ_ENEMY_DEFEAT_SHRINK)
    end

    -- 4 is the default scale, so shrink to 1/4 of regular size
    o.header.gfx.scale.x = approach_f32(o.header.gfx.scale.x, 1, 0.1, 0.1)
    if o.header.gfx.scale.x == 1 then
      cur_obj_become_tangible()
      o.oAction = WIGGLER_ACT_WALK
    end

    cur_obj_scale(o.header.gfx.scale.x)
  end
end

-- inverse of shrink action
--- @param o Object
function wiggler_custom_act_grow(o)
  -- grow bigger depending on health
  local scale = 8 - o.oHealth
  o.header.gfx.scale.x = approach_f32(o.header.gfx.scale.x, scale, 0.1, 0.1)
  if o.header.gfx.scale.x == scale then
    cur_obj_become_tangible()
    o.oAction = WIGGLER_ACT_WALK
  end

  cur_obj_scale(o.header.gfx.scale.x)
end

hook_behavior_custom(id_bhvWigglerHead, false, nil, custom_wiggler_loop)
hook_behavior_custom(id_bhvWigglerBody, false, nil, custom_wiggler_body_loop)
--hook_behavior_custom(bhvSMSRWigglerHead, false, nil, custom_wiggler_loop)

-- eyerok
local EYEROK_BOSS_ACT_ACTUALLY_SHOW_INTRO_TEXT = EYEROK_BOSS_ACT_PAUSE + 1
local EYEROK_BOSS_ACT_DIE_WITH_DIALOG = EYEROK_BOSS_ACT_PAUSE + 2
--- @param o Object
function custom_eyerok_boss_loop(o)
  -- restore text
  if o.oAction == EYEROK_BOSS_ACT_SHOW_INTRO_TEXT and o.unused1 ~= 1 then
    o.oAction = EYEROK_BOSS_ACT_ACTUALLY_SHOW_INTRO_TEXT
    o.unused1 = 1 -- doesn't sync, which is good actually
  elseif o.oAction == EYEROK_BOSS_ACT_DIE then
    o.oTimer = 0
    o.oAction = EYEROK_BOSS_ACT_DIE_WITH_DIALOG
  end

  if o.oAction == EYEROK_BOSS_ACT_ACTUALLY_SHOW_INTRO_TEXT then
    local m = gMarioStates[0]
    if should_start_or_continue_dialog(m, o) ~= 0 then
      if m.action ~= ACT_READING_NPC_DIALOG then set_mario_action(m, ACT_READING_NPC_DIALOG, 0) end
      if cutscene_object_with_dialog(CUTSCENE_DIALOG, o, gBehaviorValues.dialogs.EyerokIntroDialog) ~= 0 then
        o.oAction = EYEROK_BOSS_ACT_FIGHT
      else
        o.oTimer = o.oTimer - 1
      end
    end
  elseif o.oAction == EYEROK_BOSS_ACT_DIE_WITH_DIALOG then
    local m = gMarioStates[0]
    if (o.oTimer == 60) then
      if should_start_or_continue_dialog(m, o) ~= 0 then
        if m.action ~= ACT_READING_NPC_DIALOG then set_mario_action(m, ACT_READING_NPC_DIALOG, 0) end
        if cutscene_object_with_dialog(CUTSCENE_DIALOG, o, gBehaviorValues.dialogs.EyerokDefeatedDialog) ~= 0 then
          local starPos = gLevelValues.starPositions.EyerockStarPos
          spawn_default_star(starPos.x, starPos.y, starPos.z)
        else
          o.oTimer = o.oTimer - 1
        end
      else
        o.oTimer = o.oTimer - 1
      end
    elseif (o.oTimer > 120) then
      stop_background_music(SEQ_EVENT_BOSS)
      obj_mark_for_deletion(o)
    end
  end
  --djui_chat_message_create(tostring(o.parentObj.oEyerokBossActiveHand))
end

--- @param o Object
function custom_eyerok_hand_loop(o)
  if o.oAction == EYEROK_HAND_ACT_SHOW_EYE then -- always push towards mario
    local player = nearest_player_to_object(o)
    local angleToPlayer = obj_angle_to_object(o, player) or 0
    clamp(angleToPlayer, -0x3000, 0x3000)
    o.oMoveAngleYaw = angleToPlayer
    o.oForwardVel = 60
    if o.oTimer == 0 then -- recreation came in handy!
      o.oPosY = o.oPosY + 30
      spawn_mr_i_particle(o)
      spawn_mr_i_particle(o, o.oMoveAngleYaw - 0x1000)
      spawn_mr_i_particle(o, o.oMoveAngleYaw + 0x1000)
      o.oPosY = o.oPosY - 30
    end

    if (o.parentObj) then
      if o.parentObj.oEyerokBossNumHands == 2 then -- don't do it twice
        obj_face_yaw_approach(o.oMoveAngleYaw, 0x800)
      end

      if (o.oTimer > 10 and ((player and o.oPosZ - player.oPosZ > 0) or (o.oMoveFlags & OBJ_MOVE_HIT_EDGE) ~= 0
            or (o.oMoveFlags & OBJ_MOVE_HIT_WALL) ~= 0)) then -- also fixes bug
        --o.parentObj.oEyerokBossActiveHand = 0
        cur_obj_change_action(EYEROK_HAND_ACT_CLOSE)
        --o.parentObj.oEyerokBossActiveHand = o.oBehParams2ndByte
        o.oForwardVel = 0
      end
    end
  elseif o.oAction == EYEROK_HAND_ACT_SMASH then
    if o.oTimer < 15 then                                                             -- faster pound
      o.oTimer = 15
    elseif (o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND) ~= 0 and (o.oGravity < -3.0) then -- shockwave
      set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)
      spawn_non_sync_object(
        id_bhvBowserShockWave,
        E_MODEL_BOWSER_WAVE,
        o.oPosX, o.oFloorHeight, o.oPosZ,
        nil
      )
      o.oGravity = -3.0
    end
  elseif o.oAction == EYEROK_HAND_ACT_DOUBLE_POUND then
    if (o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND) ~= 0 then
      if o.unused1 == nil or o.unused1 == 0 then -- shockwave
        set_camera_shake_from_hit(SHAKE_LARGE_DAMAGE)
        spawn_non_sync_object(
          id_bhvBowserShockWave,
          E_MODEL_BOWSER_WAVE,
          o.oPosX, o.oFloorHeight, o.oPosZ,
          nil
        )
        o.unused1 = 30
      end
    elseif o.unused1 == nil then
      o.unused1 = 0
    elseif o.unused1 > 0 then
      o.unused1 = o.unused1 - 1
    end
  elseif o.oAction == EYEROK_HAND_ACT_FIST_SWEEP and o.oForwardVel < 10 then -- faster sweep
    o.oForwardVel = 10
  elseif o.oAction == EYEROK_HAND_ACT_CLOSE and o.oTimer > 60 then           -- fix softlock
    cur_obj_change_action(EYEROK_HAND_ACT_IDLE)
    o.parentObj.oEyerokBossUnk1AC = 0
  elseif o.oAction == EYEROK_HAND_ACT_BEGIN_DOUBLE_POUND and o.oTimer > 150 then -- fix another softlock
    o.parentObj.oEyerokBossActiveHand = o.oBehParams2ndByte
    o.oTimer = 0
    cur_obj_change_action(EYEROK_HAND_ACT_DOUBLE_POUND)
  elseif o.oAction == EYEROK_HAND_ACT_RETREAT and o.oPosY < o.oHomeY then                        -- fix other softlock
    o.oPosY = o.oHomeY
  elseif o.oAction == EYEROK_HAND_ACT_FIST_PUSH and (o.oMoveFlags & OBJ_MOVE_HIT_WALL) ~= 0 then -- yes there's more softlocks
    o.oAction = EYEROK_HAND_ACT_FIST_SWEEP
    o.oForwardVel = 0
    o.oMoveAngleYaw = 0x4000
  end
end

hook_behavior_custom(id_bhvEyerokBoss, false, nil, custom_eyerok_boss_loop)
hook_behavior_custom(id_bhvEyerokHand, false, nil, custom_eyerok_hand_loop)

-- fire piranha plants
function custom_plant_init(o)
  o.oHealth = 3
end

---@param o Object
function custom_plant_loop(o)
  if o.oBehParams >> 16 == 0 then return end

  if o.oAction == FIRE_PIRANHA_PLANT_ACT_GROW then
    local m = nearest_mario_state_to_object(o)
    if m and determine_interaction(m, o) & INT_GROUND_POUND ~= 0 then
      cur_obj_play_sound_1(SOUND_MENU_CAMERA_BUZZ)
      o.oAction = FIRE_PIRANHA_PLANT_ACT_HIDE -- hide when ground pounding
    end

    o.oFirePiranhaPlantScale = approach_f32(o.oFirePiranhaPlantScale, o.oFirePiranhaPlantNeutralScale,
      0.04 * o.oFirePiranhaPlantNeutralScale, 0)                                                                                 -- grow faster
    if o.oFirePiranhaPlantScale == o.oFirePiranhaPlantNeutralScale and obj_is_rendering_enabled() and (o.oTimer % 20) == 15 then -- spit a lot of fire
      cur_obj_init_animation_and_anim_frame(4, 56)
      cur_obj_play_sound_2(SOUND_OBJ_FLAME_BLOWN);
      obj_spit_fire(0, 30 * o.oFirePiranhaPlantNeutralScale,
        (140 * o.oFirePiranhaPlantNeutralScale),
        2.5 * o.oFirePiranhaPlantNeutralScale, 0x90, -- red flame model
        20, 15, 0x1000)
    end
  elseif o.oAction == FIRE_PIRANHA_PLANT_ACT_HIDE then
    if o.oTimer == 1 then -- less predictable
      o.oTimer = math.random(1, 70)
    end
    
    if o.oFirePiranhaPlantDeathSpinTimer == 10 and o.oHealth ~= 0 then -- survive hit (actually works in vanilla, but its made more clear they haven't died here)
      --djui_popup_create(tostring(o.oHealth), 1)
      cur_obj_play_sound_2(SOUND_OBJ2_PIRANHA_PLANT_BITE)
      o.oFirePiranhaPlantDeathSpinTimer = 0
    else
      local m = nearest_mario_state_to_object(o)
      if m and (dist_between_objects(o, m.marioObj) < 200 or determine_interaction(m, o) & INT_GROUND_POUND ~= 0) then
        o.oTimer = 0 -- don't appear when ground pounding or close to (fixes cheese sherbie discovered)
      end
    end
  end
end

hook_behavior_custom(id_bhvFirePiranhaPlant, false, custom_plant_init, custom_plant_loop)

---@param o Object
function custom_piano_init(o)
  o.oFlags = o.oFlags |
      (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
  cur_obj_update_floor()
  o.oAnimations = gObjectAnimations.mad_piano_seg5_anims_05009B14
  o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 40, -4, -0.5, 10, 10, 0
  o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
  o.oMoveAngleYaw = o.oMoveAngleYaw + 0x4000
  o.oFaceAngleYaw = o.oMoveAngleYaw
  if IS_VANILLA then
    o.oInteractType = INTERACT_BOUNCE_TOP
    o.oDamageOrCoinValue = 8
    o.oHealth = 99
  end
  bhv_init_room()
  network_init_object(o, false, { 'activeFlags' })
end

---@param o Object
function custom_piano_loop(o)
  if (not IS_VANILLA) then return bhv_mad_piano_update() end

  local m = gMarioStates[0]
  if o.oAction == 0 then -- wait for start
    cur_obj_disable_rendering_and_become_intangible(o)
    if (o.activeFlags & ACTIVE_FLAG_IN_DIFFERENT_ROOM) == 0 and o.oTimer > 5 then
      cur_obj_change_action(1)
      play_sound_with_freq_scale(SOUND_OBJ_BOWSER_LAUGH, m.marioObj.header.gfx.cameraToObject, 3.5) -- emulates the "creepy laugh" bug
      cur_obj_scale(0)
      o.oAnimState = math.random(4, 10)                                                             -- rooms to change before summoning
    end
  elseif o.oAction == 1 then                                                                        -- count down doors
    -- go to roof if outside
    cur_obj_disable_rendering_and_become_intangible(o)
    cur_obj_scale(0)
    if m.currentRoom == 13 then
      cur_obj_change_action(3)
      o.oPosX, o.oPosY, o.oPosZ = 666, 3277, 250
      o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
      o.oRoom = 13
      return
    end

    if (m.action == ACT_PULLING_DOOR or m.action == ACT_PUSHING_DOOR) and m.usedObj then
      --djui_chat_message_create(tostring(m.currentRoom))
      if o.oAnimState == 0 and m.actionTimer <= 50 then
        cur_obj_init_animation_with_sound(0)
        cur_obj_scale(1)
        o.oRoom = m.currentRoom
        local yaw = m.faceAngle.y
        if m.action == ACT_PUSHING_DOOR then
          yaw = limit_angle(yaw + 0x8000)
        end
        o.oPosX, o.oPosY, o.oPosZ = m.usedObj.oPosX + sins(yaw) * 700, m.usedObj.oPosY, m.usedObj.oPosZ + coss(yaw) * 700
        cur_obj_update_floor()
        if (not o.oFloor) or o.oFloorHeight < o.oPosY - 200 then
          o.oAnimState = 1 -- prevent getting instantly killed in the window room with the pit
        else
          o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
          o.oMoveAngleYaw = yaw + 0x4000
          o.oFaceAngleYaw = o.oMoveAngleYaw
        end
      end
      if m.actionTimer == 50 then
        if o.oAnimState == 0 then
          o.oAction = 2
          cur_obj_init_animation(1)
          o.oAnimState = math.random(4, 10)
        else
          o.oAnimState = o.oAnimState - 1
        end
      end
    elseif o.oAnimState == 0 and o.oTimer % 30 == 0 then
      play_sound(SOUND_OBJ_MAD_PIANO_CHOMPING,
        {
          x = m.marioObj.header.gfx.cameraToObject.x,
          y = m.marioObj.header.gfx.cameraToObject.y - 10000,
          z = m.marioObj.header.gfx.cameraToObject.z
        })
    end
  elseif o.oAction == 2 then -- attack
    cur_obj_enable_rendering_and_become_tangible(o)
    cur_obj_scale(1)
    o.oInteractType = INTERACT_BOUNCE_TOP
    cur_obj_update_floor_and_walls();
    cur_obj_init_animation_with_sound(1);
    cur_obj_play_sound_at_anim_range(0, 0, SOUND_OBJ_MAD_PIANO_CHOMPING);

    local angleToPlayer = obj_angle_to_object(o, m.marioObj)
    local distanceToPlayer = dist_between_objects(o, m.marioObj)
    if (distanceToPlayer < 1000) then
      o.oTimer = 0
    end

    if (o.oTimer > 80 and cur_obj_check_if_near_animation_end()) or (o.activeFlags & ACTIVE_FLAG_IN_DIFFERENT_ROOM) ~= 0 then
      o.oAction = 3
      o.oForwardVel = 0
      cur_obj_become_intangible()
    else
      cur_obj_rotate_yaw_toward(angleToPlayer, 400);
      o.oForwardVel = 10
      o.oFaceAngleYaw = o.oMoveAngleYaw
    end

    if (o.oMoveFlags & (OBJ_MOVE_LANDED | OBJ_MOVE_UNDERWATER_ON_GROUND)) ~= 0 and o.oPosY + 2000 < o.oHomeY then
      obj_mark_for_deletion(o)
      obj_explode_and_spawn_coins(46, 2)
      o.oBehParams = (10 << 24) + 2
      spawn_no_exit_star(o.oHomeX, o.oHomeY + 300, o.oHomeZ)
      if o.activeFlags & ACTIVE_FLAG_DEACTIVATED == 0 then
        obj_mark_for_deletion(o)
        network_send_object(o, true)
      end
    end

    local sMadPianoHitbox = get_temp_object_hitbox()
    sMadPianoHitbox.interactType = INTERACT_BOUNCE_TOP
    sMadPianoHitbox.downOffset = 0
    sMadPianoHitbox.damageOrCoinValue = 8
    sMadPianoHitbox.health = 99
    sMadPianoHitbox.numLootCoins = 5
    sMadPianoHitbox.radius = 200
    sMadPianoHitbox.height = 150
    sMadPianoHitbox.hurtboxRadius = 200
    sMadPianoHitbox.hurtboxHeight = 150

    if obj_check_attacks(sMadPianoHitbox, o.oAction) ~= 0 then
      o.oForwardVel = -20
      o.oVelY = 40
      o.oMoveAngleYaw = angleToPlayer
      cur_obj_change_action(4)
    end
    cur_obj_move_standard(-78)
  elseif o.oAction == 3 then -- wait
    cur_obj_become_intangible()
    cur_obj_enable_rendering()
    cur_obj_scale(1)
    cur_obj_init_animation_with_sound(0)

    local distanceToPlayer = dist_between_objects(o, m.marioObj)
    if distanceToPlayer < 500 then
      o.oAction = 2
      cur_obj_become_tangible()
      return
    end

    if (o.oMoveFlags & (OBJ_MOVE_LANDED | OBJ_MOVE_UNDERWATER_ON_GROUND)) ~= 0 and o.oPosY + 2000 < o.oHomeY then
      obj_mark_for_deletion(o)
      obj_explode_and_spawn_coins(46, 2)
      o.oBehParams = (10 << 24) + 2
      spawn_no_exit_star(o.oHomeX, o.oHomeY + 300, o.oHomeZ)
      if o.activeFlags & ACTIVE_FLAG_DEACTIVATED == 0 then
        obj_mark_for_deletion(o)
        network_send_object(o, true)
      end
    end

    if (o.activeFlags & ACTIVE_FLAG_IN_DIFFERENT_ROOM) ~= 0 and o.oTimer > 5 then
      cur_obj_change_action(1)
      cur_obj_scale(0)
      return
    end
  elseif o.oAction == 4 then -- hurt
    o.oInteractStatus = 0
    cur_obj_become_intangible()
    cur_obj_enable_rendering()
    cur_obj_scale(1)
    cur_obj_init_animation_with_sound(0)
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(100)
    if (o.oMoveFlags & (OBJ_MOVE_LANDED | OBJ_MOVE_UNDERWATER_ON_GROUND)) ~= 0 then
      if o.oPosY + 2000 < o.oHomeY then
        obj_mark_for_deletion(o)
        obj_explode_and_spawn_coins(46, 2)
        o.oBehParams = (10 << 24) + 2
        spawn_no_exit_star(o.oHomeX, o.oHomeY + 300, o.oHomeZ)
        if o.activeFlags & ACTIVE_FLAG_DEACTIVATED == 0 then
          obj_mark_for_deletion(o)
          network_send_object(o, true)
        end
      elseif o.oFloor and o.oFloor.normal.y <= 0.8 then
        o.oForwardVel = -20
        o.oVelY = 40
        local angle = atan2s(o.oFloor.normal.z, o.oFloor.normal.x) + 0x8000
        cur_obj_rotate_yaw_toward(angle, 0x2000)
      else
        cur_obj_change_action(2)
      end
    elseif (o.activeFlags & ACTIVE_FLAG_IN_DIFFERENT_ROOM) ~= 0 and o.oTimer > 5 then
      cur_obj_change_action(1)
      cur_obj_scale(0)
      return
    end
  end
end

hook_behavior_custom(id_bhvMadPiano, true, custom_piano_init, custom_piano_loop)

-- knock over chair immediately
function custom_chair_loop(o)
  if o.parentObj ~= o then
    o.oFaceAnglePitch = -0x4000
  end
end

hook_behavior_custom(id_bhvHauntedChair, false, nil, custom_chair_loop)
