-- Contains all npc data, since the cross-level stuff gets complex
-- Also handles the boss intro

IS_VANILLA = true
math.randomseed(get_time())

local specialLocation = {
  [1] = { -2800, 0, -4600 },
  [2] = { -2500, 1792, -500 },
  [3] = { 700, 942, 2700, 2 },
  --[3] = {6240, -5119, 2900},
  [4] = { -5500, -4812, -7560, 2 },
  [5] = { -2196, 819, -63 },
  --[5] = { 3330, 818, -325 },
  [6] = { 6800, -357, -1590 },
  [7] = { 6321, 640, -6570 },
  [8] = { 30, 1874, 2800, 2 },
  [9] = { 1664, 929, 605, 2 },
  [10] = { -1420, 0, -150, 2 },
  [11] = { 1720, -1407, 1720, 2 },
  [12] = { -7290, -1763, -4387, 4 },
  [13] = { 4030, -511, 2600 },
  [14] = { 0, 4475, 0 },
  [15] = { 3387, 4276, -2135 },
}

for i, v in pairs(gActiveMods) do
  if v.enabled and v.incompatible == "romhack" then
    IS_VANILLA = false
    specialLocation = {}
    break
  end
end

if _G.bossRushOn then
  IS_VANILLA = true
end

if IS_VANILLA then
  vanilla_dialog_replace()
end

local didStartingSetup = false
local FETCH_QUEST_LENGTH = 4
local FETCH_COURSE_MAX = 15
gGlobalSyncTable.babyCourse = math.random(1, FETCH_COURSE_MAX)
--djui_popup_create(tostring(gGlobalSyncTable.babyCourse), 1)
gGlobalSyncTable.fetchQuestProgress = 0
gGlobalSyncTable.fetchQuestCourse = 0

function reshuffle_command(msg)
  if not (network_is_server() or network_is_moderator()) then
    djui_chat_message_create("You can't get ye flask! (You aren't moderator)")
    return true
  end
  --[[if msg and tonumber(msg) and tonumber(msg) >= 1 then
    FETCH_COURSE_MAX = math.floor(tonumber(msg))
  end]]
  gGlobalSyncTable.babyCourse = math.random(1, FETCH_COURSE_MAX)
  if gGlobalSyncTable.fetchQuestCourse ~= 0 then
    gGlobalSyncTable.fetchQuestCourse = math.random(1, FETCH_COURSE_MAX)
    while FETCH_COURSE_MAX ~= 1 and gGlobalSyncTable.fetchQuestCourse == gGlobalSyncTable.babyCourse do
      gGlobalSyncTable.fetchQuestCourse = math.random(1, FETCH_COURSE_MAX)
    end
  end
  if didStartingSetup then
    djui_chat_message_create("Shuffled penguin and toad locations")
  end
  return true
end

hook_chat_command("reshuffle", "- Reshuffle penguin and toad locations", reshuffle_command)

local prevLevel = 1
local prevArea = 1
local storedPeng = 0
local shellFlag = false

local boss_intro_music = 0x65
secert_boss_music = 0x66
smlua_audio_utils_replace_sequence(boss_intro_music, 27, 65, "SM64DS_BossIntro27")
smlua_audio_utils_replace_sequence(secert_boss_music, 18, 65, "sulliedGrace") -- TODO: Find out who made this

local talkedToKoopa = 0
function on_warp()
  -- hard koopa the quick is very basic
  talkedToKoopa = 0 -- decheese
  gBehaviorValues.KoopaBobAgility = gBehaviorValues.KoopaCatchupAgility
  gBehaviorValues.KoopaThiAgility = gBehaviorValues.KoopaCatchupAgility + 2.0

  if not didStartingSetup then
    if IS_VANILLA then
      hook_event(HOOK_OBJECT_SET_MODEL, on_object_set_model)
    else
      for i = 2, 15 do
        if level_is_vanilla_level(course_to_level[i]) then
          FETCH_COURSE_MAX = i - 1
          reshuffle_command()
          break
        end
      end
    end
    didStartingSetup = true
  end

  if shellFlag and IS_VANILLA then
    set_koopa_dialog(false)
    shellFlag = false
  end
  playIntroOverride = false

  local np = gNetworkPlayers[0]
  local m = gMarioStates[0]

  if _G.brutalBossResetToad then
    if save_file_get_flags() & SAVE_FLAG_COLLECTED_TOAD_STAR_3 ~= 0 then
      gGlobalSyncTable.fetchQuestProgress = 0
      gGlobalSyncTable.fetchQuestCourse = 0
      save_file_clear_flags(SAVE_FLAG_COLLECTED_TOAD_STAR_3)
    end
    _G.brutalBossResetToad = false
  end

  prevLevel = np.currLevelNum
  prevArea = np.currAreaIndex

  if storedPeng > 0 then -- breaks with dynos warp
    spawn_non_sync_object(
      id_bhvSmallPenguin,
      E_MODEL_PENGUIN,
      m.pos.x, m.pos.y, m.pos.z,
      function(peng)
        peng.oBehParams = 0xF0000

        m.heldObj = peng
        peng.oHeldState = HELD_HELD
        if m.action == ACT_IDLE then
          set_mario_action(m, ACT_HOLD_IDLE, 0)
        elseif m.action == ACT_WATER_IDLE then
          set_mario_action(m, ACT_HOLD_WATER_IDLE, 0)
        end

        storedPeng = 10
      end)
  elseif gGlobalSyncTable.babyCourse == np.currCourseNum then
    local pos = nil
    local area = nil
    if specialLocation[np.currCourseNum] ~= nil then
      pos = specialLocation[np.currCourseNum]
      area = pos[4] or 1
    else
      local obj = obj_get_first_with_behavior_id(id_bhv1Up)
      if (obj == nil) then
        obj = obj_get_first(OBJ_LIST_GENACTOR)
      else
        obj_mark_for_deletion(obj)
      end
      if obj ~= nil then
        pos = { obj.oPosX, obj.oPosY, obj.oPosZ }
        if get_id_from_behavior(obj.behavior) == id_bhvBobombBuddyOpensCannon or get_id_from_behavior(obj.behavior) == id_bhvBobombBuddy then
          pos[1] = obj.oPosX + 100
        end
      else
        djui_chat_message_create("Couldn't spawn Tuxie. Check another area in this course, or use /reshuffle.")
      end
    end

    if (area == nil or np.currAreaIndex == area) and pos ~= nil then
      play_sound(SOUND_OBJ2_BABY_PENGUIN_YELL, m.marioObj.header.gfx.cameraToObject)
      spawn_non_sync_object(
        id_bhvSmallPenguin,
        E_MODEL_PENGUIN,
        pos[1], pos[2], pos[3],
        function(peng)
          peng.oBehParams = 0xF0000
          peng.oFaceAnglePitch = 0
          peng.oFaceAngleRoll = 0
        end)
    end
  end

  -- fetch quest stuff
  if gGlobalSyncTable.fetchQuestProgress > 0 and gGlobalSyncTable.fetchQuestProgress < FETCH_QUEST_LENGTH and gGlobalSyncTable.fetchQuestCourse == np.currCourseNum then
    local pos = nil
    local area = nil
    if specialLocation[np.currCourseNum] ~= nil then
      pos = specialLocation[np.currCourseNum]
      area = pos[4] or 1
    else
      local obj = obj_get_first_with_behavior_id(id_bhv1Up)
      if (obj == nil) then
        obj = obj_get_first(OBJ_LIST_GENACTOR)
      else
        obj_mark_for_deletion(obj)
      end
      if obj then
        pos = { obj.oPosX, obj.oPosY, obj.oPosZ }
        if get_id_from_behavior(obj.behavior) == id_bhvBobombBuddyOpensCannon or get_id_from_behavior(obj.behavior) == id_bhvBobombBuddy then
          pos[1] = obj.oPosX + 100
        end
      else
        djui_chat_message_create("Couldn't spawn Toad. Check another area in this course, or use /reshuffle.")
      end
    end

    if (area == nil or np.currAreaIndex == area) and pos ~= nil then
      play_sound(SOUND_MENU_YOSHI_GAIN_LIVES, m.marioObj.header.gfx.cameraToObject)
      spawn_non_sync_object(
        id_bhvToadMessage,
        E_MODEL_TOAD,
        pos[1], pos[2], pos[3],
        function(toad)
          toad.oBehParams = gBehaviorValues.dialogs.ToadStar3Dialog << 24
          toad.oBehParams2ndByte = 0xFF -- flag toad
          toad.oFaceAngleRoll = 0
          toad.oFaceAnglePitch = 0
        end)
    end
  end
end

hook_event(HOOK_ON_LEVEL_INIT, on_warp)

local play_intro_music = {
  [gBehaviorValues.dialogs.KingBobombIntroDialog] = 1,
  [gBehaviorValues.dialogs.KingWhompDialog] = 1,
  [gBehaviorValues.dialogs.WigglerDialog] = 1,
  [gBehaviorValues.dialogs.EyerokIntroDialog] = 1,
}
local playingIntro = false
playIntroOverride = false

---@param m MarioState
function mario_update(m)
  if m.action == ACT_GETTING_BLOWN then
    m.actionTimer = 0 -- prevent losing cap from wind (yes, that's the name of the action)
  elseif m.action == ACT_EATEN_BY_BUBBA then
    -- make mario invisible during this action
    m.marioObj.header.gfx.node.flags = m.marioObj.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
  end

  if m.playerIndex ~= 0 then return end

  local dialog_id = get_dialog_id()
  if play_intro_music[dialog_id] or (playIntroOverride and get_dialog_box_state() ~= 0) then
    if not playingIntro then
      seq_player_unlower_volume(SEQ_PLAYER_LEVEL, 0)
      play_music(SEQ_PLAYER_LEVEL, boss_intro_music, 0)
      playingIntro = true
    end
  elseif playingIntro then
    stop_background_music(boss_intro_music)
    playingIntro = false
  end

  if dialog_id == gBehaviorValues.dialogs.KoopaQuickBobStartDialog or dialog_id == gBehaviorValues.dialogs.KoopaQuickThiStartDialog then
    talkedToKoopa = 100
  elseif talkedToKoopa ~= 0 and hud_get_value(HUD_DISPLAY_FLAGS) & HUD_DISPLAY_FLAGS_TIMER == 0 then
    talkedToKoopa = talkedToKoopa - 1
  end

  -- koopa the quick decheese (TODO: reimplement)
  if (not mhExists) and hud_get_value(HUD_DISPLAY_FLAGS) & HUD_DISPLAY_FLAGS_TIMER ~= 0 then
    local endpoint = obj_get_first_with_behavior_id(id_bhvKoopaRaceEndpoint)
    if endpoint and endpoint.oKoopaRaceEndpointRaceEnded ~= 0 and endpoint.oKoopaRaceEndpointRaceStatus == 1 and dist_between_objects(o, m.marioObj) < 400 then
      if talkedToKoopa == 0 then
        play_sound(SOUND_OBJ_KOOPA_TALK, m.marioObj.header.gfx.cameraToObject)
        djui_popup_create("Don't try to scam ME.", 1)
        endpoint.oKoopaRaceEndpointRaceStatus = -1
      end
    end
  end

  local np = gNetworkPlayers[0]
  if m.heldObj ~= nil and get_id_from_behavior(m.heldObj.behavior) == id_bhvSmallPenguin
      and m.heldObj.oBehParams == 0xF0000 then
    storedPeng = 20
  elseif storedPeng > 0 then
    storedPeng = storedPeng - 1
  end
  if np.currLevelNum == prevLevel and np.currAreaIndex ~= prevArea then
    on_warp() -- this runs for instant warps
  end
  prevLevel = np.currLevelNum
  prevArea = np.currAreaIndex
end

hook_event(HOOK_MARIO_UPDATE, mario_update)

-- allow taking objects through doors
function on_interact(m, o, type, value)
  if (type == INTERACT_WARP_DOOR or type == INTERACT_DOOR) and m.heldObj ~= nil and o.oAction == 0 and (m.action == ACT_HOLD_WALKING or m.action == ACT_HOLD_DECELERATING) then
    local actionArg = should_push_or_pull_door(m, o)
    if type == INTERACT_WARP_DOOR then actionArg = actionArg + 4 end
    local doorData = o.oBehParams >> 24
    local doorAction = ACT_PUSHING_DOOR
    local saveFlags = save_file_get_flags()
    if type == INTERACT_DOOR then
      local doorFlag = get_door_save_file_flag(o)
      if doorData > m.numStars then
        return
      elseif doorFlag ~= 0 and (saveFlags & doorFlag) == 0 then
        doorAction = ACT_UNLOCKING_STAR_DOOR
      elseif (o.oInteractionSubtype & INT_SUBTYPE_STAR_DOOR) ~= 0 then
        doorAction = ACT_ENTERING_STAR_DOOR
      end
    else
      if doorData == 1 and (saveFlags & SAVE_FLAG_UNLOCKED_UPSTAIRS_DOOR) == 0 then
        if (saveFlags & SAVE_FLAG_HAVE_KEY_2) == 0 then return end
        doorAction = ACT_UNLOCKING_KEY_DOOR
      elseif doorData == 2 and (saveFlags & SAVE_FLAG_UNLOCKED_BASEMENT_DOOR) == 0 then
        if (saveFlags & SAVE_FLAG_HAVE_KEY_1) == 0 then return end
        doorAction = ACT_UNLOCKING_KEY_DOOR
      end
    end

    if doorAction == ACT_PUSHING_DOOR and (actionArg & 0x00000001) ~= 0 then
      doorAction = ACT_PULLING_DOOR
    end

    m.interactObj = o
    m.usedObj = o
    set_mario_action(m, doorAction, actionArg)
  elseif type == INTERACT_STAR_OR_KEY then
    --log_to_console(tostring(o.oBehParams))
    if (o.oBehParams >> 24) == 10 then -- dark star
      if m.action == ACT_FALL_AFTER_STAR_GRAB then
        set_mario_action(m, ACT_FREEFALL, 0)
      end
      play_sound(SOUND_GENERAL_GRAND_STAR, o.header.gfx.cameraToObject)
      if m.playerIndex ~= 0 then return end

      --log_to_console("Dark star get!")
      local darkStarNum = o.oBehParams & 0xFF
      save_dark_star(darkStarNum)
    end
  end
end

function before_set_mario_action(m, action)
  if m.heldObj ~= nil and action == ACT_IDLE then
    return ACT_HOLD_IDLE
  elseif m.action == ACT_RIDING_SHELL_GROUND and (not shellFlag) and IS_VANILLA then
    local dispFlags = hud_get_value(HUD_DISPLAY_FLAGS)
    if dispFlags & HUD_DISPLAY_TIMER ~= 0 then
      set_koopa_dialog(true)
      shellFlag = true
    end
  end
end

hook_event(HOOK_ON_INTERACT, on_interact)
hook_event(HOOK_BEFORE_SET_MARIO_ACTION, before_set_mario_action)
-- this function wasn't exposed to lua
function should_push_or_pull_door(m, o)
  local dx = o.oPosX - m.pos.x
  local dz = o.oPosZ - m.pos.z

  local dYaw = limit_angle(o.oMoveAngleYaw - atan2s(dz, dx))

  if (dYaw >= -0x4000 and dYaw <= 0x4000) then
    return 1
  else
    return 2
  end
end

-- from extended moveset
function limit_angle(a)
  return (a + 0x8000) % 0x10000 - 0x8000
end

function is_hazard_floor(type)
  return (type == SURFACE_DEATH_PLANE or type == SURFACE_VERTICAL_WIND or type == SURFACE_INSTANT_QUICKSAND or type == SURFACE_BURNING)
end

-- mother penguin
local penguinBossActive = false
--- @param o Object
function custom_mother_init(o)
  local levelname = get_level_name(gGlobalSyncTable.babyCourse, course_to_level[gGlobalSyncTable.babyCourse], 1) or
      'ERROR'
  penguinBossActive = false
  --djui_popup_create(tostring(levelname),1)

  smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.TuxieMotherDialog, 1, 4, 30, 200,
    string.format(DIALOG_BABY, levelname))

  smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.TuxieMotherBabyWrongDialog, 1, 4, 30, 200,
    string.format(DIALOG_WRONG_BABY, levelname))

  o.oBehParams = o.oBehParams | 0xF0000

  o.oFlags = o.oFlags |
      (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
  o.oAnimations = gObjectAnimations.penguin_seg5_anims_05008B74
  local hitbox = get_temp_object_hitbox()
  hitbox.interactType = INTERACT_TEXT
  hitbox.radius = 200
  hitbox.height = 300
  obj_set_hitbox(o, hitbox)
  cur_obj_init_animation(3)
  o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 30, -4, -0.5, 0, 0, 0
  o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
  o.oIntangibleTimer = 0
  -- I would use standard sync, except it causes crashes with this object...
  if IS_VANILLA then
    network_init_object(o, false,
      { 'oPosX', 'oPosY', 'oPosZ', 'oAction', 'oPrevAction', 'oSubAction', 'oInteractStatus', 'oHeldState',
        'oMoveAngleYaw', 'oTimer', 'activeFlags', 'oIntangibleTimer', 'oBehParams', 'oHealth' })
  end
end

--- @param o Object
function custom_mother_loop(o)
  if o.oAction < 3 then
    -- don't do text action if far away (prevents activating from really far away, for example)
    if o.oAction == 0 and o.oSubAction == 1 and dist_between_objects(o, gMarioStates[0].marioObj) > 450 then
      o.oSubAction = 2
      bhv_tuxies_mother_loop()
      o.oSubAction = 1
    else
      bhv_tuxies_mother_loop()
    end

    if _G.bossRushOn and o.oAction == 0 and o.oDialogState == 1 then
      o.oAction = 1
      o.oBehParams = 0xEFFFFF
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.TuxieMotherBabyWrongDialog, 1, 4, 30, 200,
        DIALOG_MOTHER_BOSS)
    end

    if o.oAction ~= 1 and o.prevObj and (o.prevObj.oFloorType == SURFACE_DEATH_PLANE or o.prevObj.oFloorType == SURFACE_VERTICAL_WIND) and (o.prevObj.oPosY - o.prevObj.oFloorHeight < 2048) then
      -- enter boss mode
      o.oAction = 1
      o.oBehParams = 0xEFFFFF
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.TuxieMotherBabyWrongDialog, 1, 4, 30, 200,
        DIALOG_MOTHER_BOSS)

      obj_mark_for_deletion(o.prevObj)
    elseif o.oAction == 1 and o.oBehParams == 0xEFFFFF then
      if o.oDialogState == 1 then
        playIntroOverride = true
      elseif o.oDialogState == 3 then
        penguinBossActive = true
        playIntroOverride = false
        play_music(SEQ_PLAYER_LEVEL, SEQ_EVENT_BOSS, 0)
        o.oFlags = o.oFlags & ~OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW

        local hitbox = get_temp_object_hitbox()
        hitbox.damageOrCoinValue = 3
        hitbox.radius = 50
        hitbox.height = 50
        hitbox.hurtboxHeight = 60
        hitbox.hurtboxRadius = 60
        hitbox.interactType = INTERACT_BOUNCE_TOP2
        hitbox.health = 3
        o.oFlags = o.oFlags & ~OBJ_FLAG_30
        obj_set_hitbox(o, hitbox)
        o.oFlags = o.oFlags | OBJ_FLAG_HOLDABLE

        o.oHealth = 3
        o.oSubAction = 0
        cur_obj_change_action(3)
        network_send_object(o, true)
      end
    end
    return
  elseif o.oHeldState == HELD_HELD then
    cur_obj_init_animation(3)
    cur_obj_disable_rendering_and_become_intangible(o)
    local m = gMarioStates[o.heldByPlayerIndex]
    if m then
      o.oFaceAngleYaw = m.faceAngle.y
      if o.oTimer > 300 then
        mario_drop_held_object(m)
        take_damage_and_knock_back(m, o)
        o.oHeldState = HELD_DROPPED
      end
    else
      o.oHeldState = HELD_DROPPED
    end
  elseif o.oHeldState == HELD_THROWN or o.oHeldState == HELD_DROPPED then
    cur_obj_enable_rendering()
    if o.oHeldState == HELD_THROWN then
      o.oForwardVel = 50
      o.oSubAction = 0
    else
      o.oForwardVel = 0
      o.oSubAction = 1
    end
    cur_obj_change_action(5)
    o.oMoveFlags = 0
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.oHeldState = HELD_FREE
  elseif o.oAction == 3 then -- attack
    cur_obj_init_animation(1)
    local player = nearest_player_to_object(o)
    local angleToPlayer = obj_angle_to_object(o, player) or 0
    if o.oTimer < 15 then
      o.oForwardVel = 0
      o.oFaceAngleYaw = approach_s16_symmetric(o.oFaceAngleYaw, angleToPlayer, 0x300)
      o.oMoveAngleYaw = o.oFaceAngleYaw
    else
      --djui_chat_message_create(tostring(abs_angle_diff(o.oFaceAngleYaw,angleToPlayer)))
      if o.oTimer == 15 then
        o.oForwardVel = 50
      end

      if o.oSubAction == 1 then
        o.oFaceAngleYaw = approach_s16_symmetric(o.oFaceAngleYaw, angleToPlayer, 0x300)
        o.oForwardVel = o.oForwardVel - 1
        if o.oForwardVel <= 0 then
          o.oForwardVel = 0
          o.oSubAction = 0
          o.oTimer = 13
        end
      elseif o.oMoveFlags & (OBJ_MOVE_HIT_EDGE | OBJ_MOVE_HIT_WALL) ~= 0 or abs_angle_diff(o.oFaceAngleYaw, angleToPlayer) > 0x4000 then
        o.oSubAction = 1
      end
      cur_obj_play_sound_2(SOUND_MOVING_TERRAIN_SLIDE)

      if o.oForwardVel > 40 and o.oTimer % 3 == 0 and cur_obj_abs_y_dist_to_home() >= 93 then
        spawn_non_sync_object(
          id_bhvFlameLargeBurningOut,
          E_MODEL_BLUE_FLAME,
          o.oPosX, o.oFloorHeight, o.oPosZ,
          nil
        )
      end
    end
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)

    if is_hazard_floor(o.oFloorType) and o.oVelY < -10 then -- recover from falling off on own
      o.oMoveAngleYaw = o.oMoveAngleYaw + 0x8000
      o.oVelY = 30
    end

    if o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
      cur_obj_play_sound_2(SOUND_OBJ_BIG_PENGUIN_YELL)
      o.oInteractType = INTERACT_GRABBABLE

      cur_obj_change_action(4)
      network_send_object(o, true)
    end
    o.oInteractStatus = 0
  elseif o.oAction == 4 then -- hit
    o.oForwardVel = 0
    o.oIntangibleTimer = 0
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)
    if o.oTimer > 120 then
      o.oInteractType = INTERACT_BOUNCE_TOP2

      cur_obj_change_action(3)
    end
  elseif o.oAction == 5 then -- thrown
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)
  elseif o.oAction == 6 then -- shockwaves
    if o.oTimer < 15 or (o.oVelY > -50 and o.oVelY < 0) then
      o.oVelY = -50
      local m = nearest_mario_state_to_object(o)
      if m then -- teleport from above
        o.oPosX = m.pos.x
        o.oPosZ = m.pos.z
        o.oPosY = m.pos.y + 3000
        cur_obj_update_floor_and_walls()
        if is_hazard_floor(o.oFloorType) and o.oTimer > 0 then
          o.oTimer = o.oTimer - 1
        end
      end
    else
      o.oForwardVel = 0
      cur_obj_update_floor_and_walls()
      cur_obj_move_standard(-78)
      if (o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND) ~= 0 then
        cur_obj_play_sound_2(SOUND_OBJ_THWOMP)
        set_camera_shake_from_point(SHAKE_POS_SMALL, o.oPosX, o.oPosY, o.oPosZ)
        spawn_non_sync_object(
          id_bhvBowserShockWave,
          E_MODEL_BOWSER_WAVE,
          o.oPosX, o.oFloorHeight, o.oPosZ,
          nil
        )

        if o.oSubAction < (4 - o.oHealth) then
          o.oVelY = 100
          o.oSubAction = o.oSubAction + 1
        else
          o.oSubAction = 0
          o.oAction = 3
          o.oTimer = 13
        end
      end
    end
  end

  if not penguinBossActive then
    local oldHealth = o.oHealth
    penguinBossActive = true
    playIntroOverride = false
    play_music(SEQ_PLAYER_LEVEL, SEQ_EVENT_BOSS, 0)
    o.oFlags = o.oFlags & ~OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW

    local hitbox = get_temp_object_hitbox()
    hitbox.damageOrCoinValue = 3
    hitbox.radius = 50
    hitbox.height = 50
    hitbox.hurtboxHeight = 60
    hitbox.hurtboxRadius = 60
    hitbox.interactType = INTERACT_BOUNCE_TOP2
    hitbox.health = 3
    o.oFlags = o.oFlags & ~OBJ_FLAG_30
    obj_set_hitbox(o, hitbox)
    o.oFlags = o.oFlags | OBJ_FLAG_HOLDABLE
    o.oHealth = oldHealth
  end

  -- hit detection
  if (o.oAction == 4 or o.oAction == 5) and (o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND) ~= 0 and o.oHeldState == HELD_FREE then
    cur_obj_become_tangible()
    if is_hazard_floor(o.oFloorType) then
      cur_obj_play_sound_2(SOUND_GENERAL2_RIGHT_ANSWER)
      --o.oHealth = 1
      if o.oHealth > 1 then
        o.oHealth = o.oHealth - 1
        o.oSubAction = 1
        cur_obj_change_action(6)
      else
        cur_obj_disable_rendering()
        seq_player_lower_volume(SEQ_PLAYER_LEVEL, 15, 60)
        smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.TuxieMotherBabyWrongDialog, 1, 4, 30, 200,
          DIALOG_MOTHER_DEFEAT)
        cur_obj_change_action(7)
      end
    elseif o.oAction == 5 then
      cur_obj_change_action(3)
    end
  elseif o.oAction == 7 then -- dead
    local m = gMarioStates[0]
    if should_start_or_continue_dialog(m, o) ~= 0 then
      if m.action ~= ACT_READING_NPC_DIALOG then set_mario_action(m, ACT_READING_NPC_DIALOG, 0) end
      if cutscene_object_with_dialog(CUTSCENE_DIALOG, o, gBehaviorValues.dialogs.TuxieMotherBabyWrongDialog) ~= 0 then
        stop_background_music(SEQ_EVENT_BOSS)
        seq_player_unlower_volume(SEQ_PLAYER_LEVEL, 60)
        o.oBehParams = (10 << 24) + 1
        spawn_no_exit_star(o.oHomeX, o.oHomeY + 300, o.oHomeZ)
        if o.activeFlags & ACTIVE_FLAG_DEACTIVATED == 0 then
          obj_mark_for_deletion(o)
          network_send_object(o, true)
        end
        return
      end
    end
  end

  local m
  if o.oHeldState == HELD_FREE or o.heldByPlayerIndex >= MAX_PLAYERS then
    m = nearest_mario_state_to_object(o)
  else
    m = gMarioStates[o.heldByPlayerIndex]
  end
  if m.playerIndex == 0 then
    network_send_object(o, false)
  end
end

hook_behavior_custom(id_bhvTuxiesMother, true, custom_mother_init, custom_mother_loop)

-- dark star
E_MODEL_DARK_STAR = smlua_model_util_get_id("dark_star_geo")
function on_object_set_model(o, model)
  if model == 0x7A and (o.oBehParams >> 24) == 10 and obj_has_model_extended(o, E_MODEL_DARK_STAR) == 0 then
    local darkStarNum = o.oBehParams & 0xFF
    if gGlobalSyncTable.darkStarFlags & (1 << (darkStarNum - 1)) == 0 then
      obj_set_model_extended(o, E_MODEL_DARK_STAR)
    else
      obj_set_model_extended(o, E_MODEL_TRANSPARENT_STAR)
    end
  end
end

-- racing penguin
local marioLastFloorHeight = 0
--- @param o Object
function custom_race_penguin_loop(o)
  if o.oAction ~= 3 then return end

  local m = gMarioStates[0]
  if o.oForwardVel < 87 then -- just barely slower than mario
    o.oForwardVel = o.oForwardVel * 1.1
  end

  -- better cheating check
  if (o.oRacingPenguinMarioCheated == 0) and mario_is_in_air_action(m) == 0 then
    if (marioLastFloorHeight - m.pos.y) > 1000 then
      play_sound(SOUND_OBJ_BIG_PENGUIN_YELL, m.marioObj.header.gfx.cameraToObject)
      o.oRacingPenguinMarioCheated = 1
      djui_popup_create("No cheating!", 1)
    else
      marioLastFloorHeight = m.pos.y
    end
  end

  if o.oTimer > 10 then
    spawn_non_sync_object(
      id_bhvFlameLargeBurningOut,
      E_MODEL_BLUE_FLAME,
      o.oPosX, o.oPosY, o.oPosZ,
      nil
    )

    o.oTimer = 0
  end
end

hook_behavior_custom(id_bhvRacingPenguin, false, nil, custom_race_penguin_loop)
--hook_behavior_custom(bhvSMSRRacingPenguin, false, nil, custom_race_penguin_loop)

-- mips
local mipsTeleportCount = 0
local MIPS_ACT_CUSTOM_VANISH = MIPS_ACT_IDLE + 1
--- @param o Object
function custom_mips_loop(o)
  local m = nearest_mario_state_to_object(o)
  if o.oAction == MIPS_ACT_WAIT_FOR_NEARBY_MARIO and o.oBehParams2ndByte ~= 0 then
    cur_obj_change_action(MIPS_ACT_IDLE)
  elseif m and o.oBehParams2ndByte == 0 then
    o.oMipsForwardVelocity = math.max(45, m.forwardVel + 20)
  end

  if o.oAction == MIPS_ACT_CUSTOM_VANISH then
    if o.oSubAction == 0 then
      if m then
        o.oMoveAngleYaw = obj_angle_to_object(o, m.marioObj)
      end

      if o.oTimer > 10 then
        o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
        spawn_mist_particles_variable(0, 10, 30)
        o.oSubAction = 1
        cur_obj_play_sound_2(SOUND_OBJ2_BOWSER_TELEPORT)
        mipsTeleportCount = mipsTeleportCount + 1
      end
    elseif not (m and dist_between_objects(o, m.marioObj) <= 1000) then
      o.header.gfx.node.flags = o.header.gfx.node.flags & ~GRAPH_RENDER_INVISIBLE
      cur_obj_init_animation(2)
      cur_obj_change_action(MIPS_ACT_WAIT_FOR_ANIMATION_DONE)
      cur_obj_become_tangible()
      spawn_mist_particles_variable(0, 10, 30)
      cur_obj_play_sound_2(SOUND_OBJ2_BOWSER_TELEPORT)
      if mipsTeleportCount >= 10 then
        mipsTeleportCount = 0
        djui_chat_message_create("You seem to be having trouble. Try sneaking up from behind...")
      end
    end
  elseif (o.oMipsStarStatus == MIPS_STAR_STATUS_HAVENT_SPAWNED_STAR and o.oHeldState == HELD_FREE and (o.oAction == MIPS_ACT_IDLE or o.oAction == MIPS_ACT_WAIT_FOR_ANIMATION_DONE) and m and dist_between_objects(o, m.marioObj) <= 1000) then
    if m.forwardVel > 10 or (m.action & ACT_FLAG_AIR) ~= 0 or abs_angle_diff(obj_angle_to_object(o, m.marioObj), o.oFaceAngleYaw) < 0x5000 then -- require sneakup
      cur_obj_become_intangible()
      cur_obj_init_animation(2)
      cur_obj_play_sound_2(SOUND_OBJ_GOOMBA_ALERT)
      cur_obj_change_action(MIPS_ACT_CUSTOM_VANISH) -- custom action
      o.oSubAction = 0
    end
  end
  --[[if (m and m.prevAction ~= ACT_DIVE and m.action == ACT_DIVE and o.oAction == MIPS_ACT_FOLLOW_PATH and o.oPathedStartWaypoint) then
    m.prevAction = ACT_DIVE
    o.oPosX = o.oPathedStartWaypoint.pos.x
    o.oPosY = o.oPathedStartWaypoint.pos.y
    o.oPosZ = o.oPathedStartWaypoint.pos.z
    cur_obj_follow_path(0)
    cur_obj_play_sound_2(SOUND_OBJ2_BOWSER_TELEPORT)
  end]]
end

hook_behavior_custom(id_bhvMips, false, nil, custom_mips_loop)

--[[save_file_clear_flags(SAVE_FLAG_COLLECTED_TOAD_STAR_1)
save_file_clear_flags(SAVE_FLAG_COLLECTED_TOAD_STAR_2)
save_file_clear_flags(SAVE_FLAG_COLLECTED_TOAD_STAR_3)
save_file_clear_flags(SAVE_FLAG_COLLECTED_MIPS_STAR_1)
save_file_clear_flags(SAVE_FLAG_COLLECTED_MIPS_STAR_2)]]

function custom_toad_message_init(o)
  o.oFlags = o.oFlags |
      (OBJ_FLAG_PERSISTENT_RESPAWN | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
  o.oAnimations = gObjectAnimations.toad_seg6_anims_0600FB58
  cur_obj_init_animation(6)
  local hitbox = get_temp_object_hitbox()
  hitbox.interactType = INTERACT_TEXT
  hitbox.radius = 80
  hitbox.height = 100
  obj_set_hitbox(o, hitbox)
  o.oIntangibleTimer = 0
  bhv_init_room()
  bhv_toad_message_init()

  o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
  cur_obj_update_floor()
  if o.oBehParams2ndByte == 0xFF and (not IS_VANILLA) then
    local waterLevel = cur_obj_move_y_and_get_water_level(0, 0)
    if waterLevel > o.oPosY then
      o.oPosY = waterLevel + 30
      spawn_non_sync_object(
        id_bhvStaticCheckeredPlatform,
        E_MODEL_CHECKERBOARD_PLATFORM,
        o.oPosX, o.oPosY - 25, o.oPosZ,
        nil
      )
    elseif is_hazard_floor(o.oFloorType) then
      spawn_non_sync_object(
        id_bhvStaticCheckeredPlatform,
        E_MODEL_CHECKERBOARD_PLATFORM,
        o.oPosX, o.oPosY, o.oPosZ,
        nil
      )
      o.oPosY = o.oPosY + 25
    else
      o.oPosY = o.oFloorHeight
    end
  end
end

--- @param o Object
function custom_toad_message_loop(o)
  if o.oAction ~= 0 then
    sir_bup_loop(o)
    --bhv_toad_message_loop()
  elseif playIntroOverride and o.oDialogState == 3 then -- switch to sir bup
    cur_obj_change_action(1)

    local hitbox = get_temp_object_hitbox()
    hitbox.damageOrCoinValue = 3
    hitbox.radius = 40
    hitbox.height = 60
    hitbox.hurtboxHeight = 30
    hitbox.hurtboxRadius = 50
    hitbox.interactType = INTERACT_BOUNCE_TOP2
    hitbox.health = 3
    o.oHealth = 3
    o.oFlags = o.oFlags & ~OBJ_FLAG_30
    bhv_toad_message_loop()
    obj_set_hitbox(o, hitbox)
    o.oToadMessageState = 1
    o.oInteractStatus = 0
    o.oGravity = -2

    --[[local m = nearest_mario_state_to_object(o)
    if m then
      take_damage_and_knock_back(m, o)
    end]]

    play_music(SEQ_PLAYER_LEVEL, SEQ_EVENT_BOSS, 0)
    playIntroOverride = false
    return
  end

  bhv_toad_message_loop()

  if o.oToadMessageDialogId == DIALOG_136 and IS_VANILLA then -- the basement toad gives dark star hints
    if o.oDialogState == 1 and o.unused1 == 0 then
      set_dark_star_toad_dialog()
      o.unused1 = 1
    elseif o.oDialogState == 0 and o.unused1 ~= 0 then
      o.unused1 = 0
    end
    return
  end
  if o.oDialogState ~= 1 then return end

  local isToad = (gMarioStates[0].character.type == 2)
  if isToad and charSelectExists then
    local charData = charSelect.character_get_current_table()
    if charData and charData.name and not (charData.name:find("Toad") or charData.name:find("Mush")) then
      isToad = false
    end
  end

  if o.oToadMessageDialogId == gBehaviorValues.dialogs.ToadStar1Dialog then -- toad star 1 requires that the player be Toad
    if not isToad then
      o.oToadMessageDialogId = DIALOG_099 -- plays boo sound, oh well

      smlua_text_utils_dialog_replace(DIALOG_099, 1, 3, 30, 200, DIALOG_FELLOW_TOAD)
    else
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar1Dialog, 1, 3, 30, 200, DIALOG_TALK_TOAD)
    end
  elseif o.oToadMessageDialogId == DIALOG_099 then -- also toad star 1
    if isToad then
      local saveFlags = save_file_get_flags()
      if (saveFlags & SAVE_FLAG_COLLECTED_TOAD_STAR_1) == 0 then
        o.oToadMessageDialogId = gBehaviorValues.dialogs.ToadStar1Dialog

        smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar1Dialog, 1, 3, 30, 200, DIALOG_TALK_TOAD)
      end
    end
  elseif o.oToadMessageDialogId == gBehaviorValues.dialogs.ToadStar2Dialog and o.oAction ~= 4 then -- toad star 2 is a joke boss fight
    o.oToadMessageDialogId = gBehaviorValues.dialogs.ToadStar2AfterDialog
    playIntroOverride = true

    smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar2Dialog, 1, 6, 30, 200, DIALOG_BESTED)

    smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar2AfterDialog, 1, 3, 30, 200, DIALOG_SIR_BUP)
  elseif o.oToadMessageDialogId == gBehaviorValues.dialogs.ToadStar3Dialog then -- toad star 3 is a wild goose chase
    local course = gNetworkPlayers[0].currCourseNum or 0
    if gGlobalSyncTable.fetchQuestProgress == 0 then                            -- first dialog
      gGlobalSyncTable.fetchQuestProgress = 1
      gGlobalSyncTable.fetchQuestCourse = math.random(1, FETCH_COURSE_MAX)
      while FETCH_COURSE_MAX ~= 1 and gGlobalSyncTable.fetchQuestCourse == gGlobalSyncTable.babyCourse do
        gGlobalSyncTable.fetchQuestCourse = math.random(1, FETCH_COURSE_MAX)
      end

      local levelname = get_level_name(gGlobalSyncTable.fetchQuestCourse,
        course_to_level[gGlobalSyncTable.fetchQuestCourse], 1) or 'ERROR'

      o.oToadMessageDialogId = gBehaviorValues.dialogs.ToadStar3AfterDialog
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar3AfterDialog, 1, 3, 30, 200,
        string.format(DIALOG_FETCH_START, levelname))
    elseif gGlobalSyncTable.fetchQuestProgress < FETCH_QUEST_LENGTH then
      o.oToadMessageDialogId = gBehaviorValues.dialogs.ToadStar3AfterDialog
      if course == 0 then
        local levelname = get_level_name(gGlobalSyncTable.fetchQuestCourse,
          course_to_level[gGlobalSyncTable.fetchQuestCourse], 1) or 'ERROR'

        smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar3AfterDialog, 1, 3, 30, 200,
          string.format(DIALOG_FETCH_PROGRESS, levelname))
      else
        if course == gGlobalSyncTable.fetchQuestCourse then
          gGlobalSyncTable.fetchQuestProgress = gGlobalSyncTable.fetchQuestProgress + 1
          local prevCourse = gGlobalSyncTable.fetchQuestCourse
          gGlobalSyncTable.fetchQuestCourse = math.random(1, FETCH_COURSE_MAX)
          while FETCH_COURSE_MAX ~= 1 and gGlobalSyncTable.fetchQuestCourse == gGlobalSyncTable.babyCourse or gGlobalSyncTable.fetchQuestCourse == prevCourse do
            gGlobalSyncTable.fetchQuestCourse = math.random(1, FETCH_COURSE_MAX)
          end
        end

        if gGlobalSyncTable.fetchQuestProgress == FETCH_QUEST_LENGTH then
          gGlobalSyncTable.fetchQuestCourse = 0

          smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar3AfterDialog, 1, 4, 30, 200,
            DIALOG_ORIGINAL_OWNER)
        else
          set_toad_3_dialog(gGlobalSyncTable.fetchQuestCourse)
        end
      end
    elseif o.oBehParams2ndByte == 0xFF then -- in case someone talks to the second to last toad, then another person tries to talk to them
      o.oToadMessageDialogId = gBehaviorValues.dialogs.ToadStar3AfterDialog

      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar3AfterDialog, 1, 4, 30, 200, DIALOG_ORIGINAL_OWNER)
    else
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar3AfterDialog, 1, 6, 30, 200, DIALOG_GOOSE_CHASE)
    end
  end
end

local toadInvincTimer = 0
--- @param o Object
function sir_bup_loop(o)
  o.oOpacity = 255
  if toadInvincTimer % 2 == 1 then
    o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
  else
    o.header.gfx.node.flags = o.header.gfx.node.flags & ~GRAPH_RENDER_INVISIBLE
  end
  --djui_chat_message_create(string.format("0x%0x",o.oMoveFlags))

  local player = gMarioStates[0].marioObj
  local angleToPlayer = obj_angle_to_object(o, player) or 0
  local dist = dist_between_objects(o, gMarioStates[0].marioObj) or 1000

  if o.oAction == 1 then -- jump towards player
    cur_obj_rotate_yaw_toward(angleToPlayer, 0x300)
    o.oFaceAngleYaw = o.oMoveAngleYaw
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)

    if o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
      if dist < 150 then
        o.oForwardVel = approach_f32(o.oForwardVel, 0, 2, 2)
        o.oVelY = 20
        cur_obj_play_sound_2(SOUND_MARIO_YAH_WAH_HOO)
      elseif dist < 700 then
        o.oForwardVel = 20
        o.oVelY = 30
        cur_obj_play_sound_2(SOUND_MARIO_YAH_WAH_HOO)
      else
        o.oForwardVel = 40
        o.oVelY = 20
        o.oSubAction = 0
        cur_obj_change_action(5) -- switch to dives
        cur_obj_play_sound_2(SOUND_MARIO_HOOHOO)
      end
    elseif o.oMoveFlags & OBJ_MOVE_HIT_WALL ~= 0 then
      cur_obj_reflect_move_angle_off_wall()
    else
      local dist = dist_between_object_and_point(o, gMarioStates[0].pos.x, o.oPosY, gMarioStates[0].pos.z) or 1000
      cur_obj_init_animation_and_anim_frame(0, 25) -- jump animation

      if o.oVelY <= 0 and dist < 150 then
        cur_obj_play_sound_2(SOUND_MARIO_GROUND_POUND_WAH)
        cur_obj_change_action(3)
      end
    end
  elseif o.oAction == 2 then -- knockback
    cur_obj_init_animation(6)
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)

    if o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
      if o.oMoveFlags & OBJ_MOVE_LANDED ~= 0 then
        cur_obj_play_sound_2(SOUND_MARIO_OOOF2)
      end
      o.oFaceAnglePitch = -0x4000
      o.oForwardVel = 0
      if o.oTimer > 30 then
        if o.oHealth <= 1 then
          o.oTimer = 0
          cur_obj_change_action(4)
        else
          o.oHealth = o.oHealth - 1
          o.oFaceAnglePitch = 0
          o.oForwardVel = 0
          o.oVelY = 10
          cur_obj_play_sound_2(SOUND_MARIO_YAH_WAH_HOO)
          toadInvincTimer = 120
          cur_obj_change_action(1)
        end
      end
    elseif o.oMoveFlags & OBJ_MOVE_HIT_WALL ~= 0 then
      cur_obj_reflect_move_angle_off_wall()
    else
      o.oFaceAnglePitch = o.oFaceAnglePitch - 0x1000
      o.oTimer = 0
    end
  elseif o.oAction == 3 then -- ground pound
    cur_obj_init_animation(6)
    o.oVelY = -50
    o.oForwardVel = 0

    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)

    if o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
      if o.oMoveFlags & OBJ_MOVE_LANDED ~= 0 then
        cur_obj_play_sound_2(SOUND_ACTION_TERRAIN_HEAVY_LANDING)
        spawn_mist_particles()
        local m = gMarioStates[0]
        --djui_chat_message_create(tostring((m.action & ACT_FLAG_INVULNERABLE))..":"..tostring(dist))
        if dist < 150 and toadInvincTimer == 0 and m.invincTimer == 0 and (m.action & ACT_FLAG_INVULNERABLE) == 0 and m.health > 0xFF then
          take_damage_and_knock_back(m, o)
        end
      end
      if o.oTimer > 10 then
        cur_obj_change_action(1)
      end
    else
      o.oTimer = 0
    end
  elseif o.oAction == 4 then -- death
    cur_obj_init_animation_and_anim_frame(6, 0)
    if o.oTimer == 30 then
      --cur_obj_init_animation(6)
      seq_player_lower_volume(SEQ_PLAYER_LEVEL, 15, 60)

      o.oToadMessageDialogId = gBehaviorValues.dialogs.ToadStar2Dialog
      o.oToadMessageState = 4
      smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.ToadStar2AfterDialog, 1, 3, 30, 200, DIALOG_WEIRD)
    elseif o.oTimer > 30 and o.oToadMessageState ~= 4 then
      stop_background_music(SEQ_EVENT_BOSS)
      seq_player_unlower_volume(SEQ_PLAYER_LEVEL, 15)
      obj_mark_for_deletion(o)
      spawn_mist_particles()
    end

    bhv_toad_message_loop()
  elseif o.oAction == 5 then -- dive
    cur_obj_init_animation(7)
    cur_obj_rotate_yaw_toward(angleToPlayer, 0x200)

    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)

    if o.oMoveFlags & OBJ_MOVE_MASK_ON_GROUND ~= 0 then
      if o.oSubAction == 1 then
        local dist = dist_between_object_and_point(o, gMarioStates[0].pos.x, o.oPosY, gMarioStates[0].pos.z) or 1000
        if dist >= 700 then
          o.oFaceAnglePitch = 0
          o.oForwardVel = 40
          o.oVelY = 20
          o.oSubAction = 0
          cur_obj_play_sound_2(SOUND_MARIO_HOOHOO)
        else
          o.oForwardVel = 20
          o.oVelY = 40
          o.oFaceAnglePitch = 0
          cur_obj_change_action(1) -- switch to back to jumps
          cur_obj_play_sound_2(SOUND_MARIO_YAH_WAH_HOO)
        end
      else
        o.oForwardVel = 30
        o.oVelY = 20
        o.oSubAction = 1
        cur_obj_play_sound_2(SOUND_MARIO_HOOHOO)
      end
    elseif o.oMoveFlags & OBJ_MOVE_HIT_WALL ~= 0 then
      cur_obj_reflect_move_angle_off_wall()
    elseif o.oSubAction == 1 then
      o.oFaceAnglePitch = approach_f32(o.oFaceAnglePitch, 0xFFFF, 0x1500, 0x1500)
    else
      o.oFaceAnglePitch = approach_f32(o.oFaceAnglePitch, 0x4000, 0x800, 0x800)
    end
  end

  if (o.oAction ~= 4 and o.oAction ~= 2 and toadInvincTimer == 0) and (o.oInteractStatus & INT_STATUS_WAS_ATTACKED) ~= 0 then
    o.oForwardVel = -20
    o.oVelY = 40
    o.oMoveAngleYaw = angleToPlayer
    cur_obj_play_sound_2(SOUND_MARIO_DOH)
    cur_obj_change_action(2)
    cur_obj_become_intangible()
  end
  if toadInvincTimer > 0 then
    toadInvincTimer = toadInvincTimer - 1
    if toadInvincTimer == 0 then
      cur_obj_become_tangible()
    end
  end
  if o.oPosY == o.oFloorHeight and is_hazard_floor(o.oFloorType) then
    cur_obj_set_pos_to_home()
  end
  o.oInteractStatus = 0
end

hook_behavior_custom(id_bhvToadMessage, true, custom_toad_message_init, custom_toad_message_loop)

-- ukiki
function custom_ukiki_init(o)
  o.oFlags = o.oFlags |
      (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_HOLDABLE | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
  local hitbox = get_temp_object_hitbox()
  hitbox.interactType = INTERACT_GRABBABLE
  hitbox.radius = 40
  hitbox.height = 40
  obj_set_hitbox(o, hitbox)
  o.oInteractionSubtype = INT_SUBTYPE_HOLDABLE_NPC
  o.oIntangibleTimer = 0
  cur_obj_update_floor()
  o.oAnimations = gObjectAnimations.ukiki_seg5_anims_05015784
  cur_obj_init_animation(0)
  o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 30, -4, -0.5, 0, 0, 0
  o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
  bhv_ukiki_init()
end

function custom_ukiki_loop(o)
  if o.oHeldState ~= HELD_FREE then
    if o.heldByPlayerIndex == 0 then
      if o.oBehParams2ndByte == UKIKI_CAP then
        if o.oUkikiHasCap & UKIKI_CAP_ON ~= 0 then
          o.oUkikiHasCap = o.oUkikiHasCap & ~UKIKI_CAP_ON
          o.oUkikiTextState = UKIKI_TEXT_HAS_CAP
        elseif o.oUkikiTextState == UKIKI_TEXT_DEFAULT or o.oUkikiTextState == UKIKI_TEXT_STOLE_CAP then
          o.oUkikiTextState = UKIKI_TEXT_STOLE_CAP
          o.oInteractionSubtype = o.oInteractionSubtype | INT_SUBTYPE_DROP_IMMEDIATELY
        end
      elseif o.oBehParams ~= 0xFF then
        smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.UkikiHeldDialog, 1, 4, 30, 200,
          DIALOG_UKIKI_FRIEND)
        if o.oUkikiTextState == UKIKI_TEXT_DO_NOT_LET_GO then -- punish
          o.oDamageOrCoinValue = 2
          take_damage_and_knock_back(gMarioStates[0], o)
          o.oUkikiTextState = UKIKI_TEXT_DEFAULT
        end
      else
        smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.UkikiHeldDialog, 1, 4, 30, 200,
          DIALOG_UKIKI_OPEN)
      end
    end
    if o.oHeldState ~= HELD_HELD then
      if o.oBehParams2ndByte == UKIKI_CAP then
        cur_obj_change_action(6) -- unused turn action, used as idle
        local other = obj_get_first_with_behavior_id(id_bhvUkiki)
        if other == o then other = obj_get_next_with_same_behavior_id(other) end
        if other and other.oBehParams2ndByte == UKIKI_CAGE then
          o.oBehParams = 0xFF
          o.oBehParams2ndByte = UKIKI_CAGE
          o.oUkikiTextState = UKIKI_TEXT_DEFAULT
          o.oHomeY = other.oHomeY
          o.oInteractionSubtype = 0 -- needed to make this even possible
        end
      elseif o.oBehParams ~= 0xFF then
        if o.oUkikiTextState == UKIKI_TEXT_GO_TO_CAGE then
          cur_obj_change_action(6) -- unused turn action, used as idle
        end
      end
    end
    if o.oHeldState == HELD_THROWN then
      cur_obj_enable_rendering_2()
      o.header.gfx.node.flags = o.header.gfx.node.flags & ~GRAPH_RENDER_INVISIBLE
      o.oHeldState = 0
      o.oForwardVel = 50
      o.oVelY = 30
      o.oSubAction = 0
      cur_obj_change_action(9) -- thrown action
    else
      bhv_ukiki_loop()
    end
  elseif o.oAction == 8 then -- custom action, wait for dialog end
    local m = nearest_mario_state_to_object(o)
    cur_obj_init_animation_with_sound(UKIKI_ANIM_JUMP_CLAP)
    if m and m.playerIndex == 0 and should_start_or_continue_dialog(m, o) ~= 0 then
      if m.action ~= ACT_READING_NPC_DIALOG then set_mario_action(m, ACT_READING_NPC_DIALOG, 0) end
      if cutscene_object_with_dialog(CUTSCENE_DIALOG, o, gBehaviorValues.dialogs.UkikiCapStealDialog) ~= 0 then
        cur_obj_change_action(1)
        network_send_object(o, true)
      end
    end
  elseif o.oAction == 9 then -- custom thrown action
    cur_obj_become_intangible()
    cur_obj_move_standard(20)
    cur_obj_update_floor_and_walls()
    if (o.oSubAction == 0) then
      cur_obj_init_animation_with_sound(UKIKI_ANIM_HELD)
      if (o.oMoveFlags & (OBJ_MOVE_MASK_ON_GROUND | OBJ_MOVE_UNDERWATER_ON_GROUND)) ~= 0 or o.oTimer >= 300 then
        o.oSubAction = 1
        o.oVelY = 0
      end
    else
      o.oForwardVel = 0
      cur_obj_init_animation_with_sound(UKIKI_ANIM_LAND);
      cur_obj_become_tangible()

      if (cur_obj_check_if_near_animation_end() ~= 0) then
        cur_obj_change_action(6) -- unused turn action, used as idle
      end
    end
  elseif o.oAction == 1 then -- custom run act (steal star)
    local m = nearest_mario_state_to_object(o)
    if not m then return bhv_ukiki_loop() end

    cur_obj_update_floor_and_walls()
    local dist = dist_between_objects(o, m.marioObj)
    local goalYaw = obj_angle_to_object(o, m.marioObj)
    local fleeMario = false

    if o.oUkikiHasCap & UKIKI_CAP_ON ~= 0 or o.oBehParams2ndByte ~= UKIKI_CAP then
      fleeMario = true
      goalYaw = limit_angle(goalYaw + 0x8000)
    end

    cur_obj_init_animation_with_sound(UKIKI_ANIM_RUN)
    cur_obj_rotate_yaw_toward(goalYaw, 0x800);
    cur_obj_set_vel_from_mario_vel(m, 20, 1.5)
    if m.playerIndex == 0 and m.numStars ~= 0 and (not fleeMario) and mario_ready_to_speak(m) ~= 0 and dist <= 200 then
      cur_obj_change_action(8)

      -- remove random star
      local file = get_current_save_file_num() - 1
      local courseStarFlags = {}
      local validCourses = {}
      for i = 0, COURSE_MAX - 1 do
        local starFlags = save_file_get_star_flags(file, i - 1)
        if starFlags & ~0x80 ~= 0 then
          courseStarFlags[i] = starFlags
          table.insert(validCourses, i)
        end
      end
      if #validCourses ~= 0 then
        local course = validCourses[math.random(1, #validCourses)]
        local starFlags = courseStarFlags[course]
        local starNum = math.random(0, 6)
        for i = 0, 6 do
          if starFlags & (1 << starNum) ~= 0 then
            save_file_remove_star_flags(file, course - 1, (1 << starNum))
            starNum = starNum + 1
            break
          end
          starNum = starNum + 1
          if starNum == 7 then starNum = 0 end
        end
        update_all_mario_stars()
        network_send(true, { id = PACKET_UPDATE_STARS })

        --djui_chat_message_create(get_star_name(course, starNum))
        local star = spawn_sync_object(id_bhvUkikiHoldStar, E_MODEL_STAR, o.oPosX, o.oPosY, o.oPosZ, nil)
        star.parentObj = o
        star.oBobombBlinkTimer = course
        star.oBobombFuseTimer = starNum
        network_send_object(o, true)
      end

      o.oUkikiHasCap = o.oUkikiHasCap | UKIKI_CAP_ON
      return
    end

    if fleeMario then
      if (o.oMoveFlags & OBJ_MOVE_HIT_WALL) ~= 0 then
        o.oAction = UKIKI_ACT_JUMP
        o.oMoveAngleYaw = o.oWallAngle
      elseif (o.oMoveFlags & OBJ_MOVE_HIT_EDGE) ~= 0 then
        o.oAction = UKIKI_ACT_JUMP
        o.oMoveAngleYaw = o.oMoveAngleYaw + 0x8000
      elseif (dist > 1000) then
        o.oAction = UKIKI_ACT_TURN_TO_MARIO
      end
    end
    cur_obj_move_standard(-20)
  elseif o.oUkikiHasCap & UKIKI_CAP_ON ~= 0 then -- spoof no cap state
    local saveFlags = gMarioStates[0].flags
    gMarioStates[0].flags = 0
    bhv_ukiki_loop()
    gMarioStates[0].flags = saveFlags
    o.oAnimState = UKIKI_ANIM_STATE_DEFAULT
  else
    bhv_ukiki_loop()
    if is_hazard_floor(o.oFloorType) then
      cur_obj_set_pos_to_home()
    end
  end
end

id_bhvUkiki = hook_behavior_custom(id_bhvUkiki, true, custom_ukiki_init, custom_ukiki_loop)

function ukiki_hold_star_init(o)
  o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
  o.oFaceAngleYaw = 0
  o.oFaceAnglePitch = 0
  o.oFaceAngleRoll = 0
  network_init_object(o, false, {
    'activeFlags',
    'oBobombBlinkTimer',
    'oBobombFuseTimer',
  })
end

function ukiki_hold_star_loop(o)
  o.oFaceAngleYaw = o.oFaceAngleYaw + 0x500
  o.oAnimState = o.oBobombBlinkTimer -- this makes the star's color match what course it's from in OMM

  -- there's no way to sync the parent object, so just check for the nearest ukiki
  if o.parentObj == nil or o.parentObj == o then
    o.parentObj = obj_get_nearest_object_with_behavior_id(o, id_bhvUkiki)
  end

  if o.oBobombFuseTimer ~= 0 and (o.parentObj == nil or o.parentObj == o or (o.parentObj.oUkikiHasCap & UKIKI_CAP_ON == 0 and o.parentObj.oHeldState ~= HELD_HELD)) then
    -- return star
    local file = get_current_save_file_num() - 1
    save_file_set_star_flags(file, o.oBobombBlinkTimer - 1, (1 << (o.oBobombFuseTimer - 1)))
    cur_obj_play_sound_2(SOUND_MENU_STAR_SOUND)
    spawn_mist_particles()
    update_all_mario_stars()
    network_send(true, { id = PACKET_UPDATE_STARS })
    obj_mark_for_deletion(o)
  end
  if o.parentObj and o.parentObj ~= o then
    o.oPosX = o.parentObj.oPosX
    o.oPosY = o.parentObj.oPosY + 200
    o.oPosZ = o.parentObj.oPosZ
  end
end

id_bhvUkikiHoldStar = hook_behavior(nil, OBJ_LIST_DEFAULT, true, ukiki_hold_star_init, ukiki_hold_star_loop,
  "bhvUkikiHoldStar")

-- klepto (is klepto an npc? whatever)
function custom_klepto_loop(o)
  local course = gNetworkPlayers[0].currCourseNum
  local file = get_current_save_file_num() - 1
  -- give klepto the star if it isn't collected
  if gNetworkPlayers[0].currAreaSyncValid and o.oAnimState == KLEPTO_ANIM_STATE_HOLDING_NOTHING and o.oBehParams2ndByte == 0 and save_file_get_star_flags(file, course - 1) & 1 == 0 then
    o.oBehParams2ndByte = 1
    o.oAnimState = KLEPTO_ANIM_STATE_HOLDING_STAR
    if o.oAction == KLEPTO_ACT_WAIT_FOR_MARIO or o.oAction == KLEPTO_ACT_TURN_TOWARD_MARIO then
      cur_obj_change_action(0)
      o.oFlags = o.oFlags | 0x8
    end
  end

  if o.oAction == KLEPTO_ACT_WAIT_FOR_MARIO then -- relentlessly attack mario
    o.oAction = KLEPTO_ACT_TURN_TOWARD_MARIO
    o.oFlags = o.oFlags & ~0x8
  elseif o.oAction == KLEPTO_ACT_CIRCLE_TARGET_HOLDING then -- always fly around
    o.oKleptoTimeUntilTargetChange = 0
  elseif o.oAnimState == KLEPTO_ANIM_STATE_HOLDING_CAP then -- take star instead of cap
    local m = gMarioStates[0]
    o.oAnimState = KLEPTO_ANIM_STATE_HOLDING_NOTHING
    m.cap = 0
    m.flags = m.flags | MARIO_CAP_ON_HEAD | MARIO_NORMAL_CAP
    save_file_clear_flags(SAVE_FLAG_CAP_ON_KLEPTO | SAVE_FLAG_CAP_ON_UKIKI)
    if o.oBehParams2ndByte == 0 then -- we can't spawn a star twice, so only steal once
      if save_file_get_star_flags(file, course - 1) & 1 ~= 0 then
        o.oBehParams2ndByte = 1
        o.oAnimState = KLEPTO_ANIM_STATE_HOLDING_STAR
        save_file_remove_star_flags(file, course - 1, 1)
        update_all_mario_stars()
        network_send(true, { id = PACKET_UPDATE_STARS })
        network_send_object(o, true)
      end
    end
  end
end

hook_behavior(id_bhvKlepto, OBJ_LIST_GENACTOR, false, nil, custom_klepto_loop, "bhvKlepto")

-- snowman's head
function custom_sl_wind_init(o)
  o.oFlags = o.oFlags |
      (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
  o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
end

function custom_sl_wind_loop(o)
  if o.oSubAction ~= SL_SNOWMAN_WIND_ACT_BLOWING then
    bhv_sl_snowman_wind_loop()
    if o.oTimer == 1 then
      o.oSLSnowmanWindOriginalYaw = o.oMoveAngleYaw
    end
  end

  if o.oSubAction ~= SL_SNOWMAN_WIND_ACT_TALKING then -- blow when idle to annoy speedrunners
    local player = nearest_player_to_object(o)
    local distanceToPlayer = dist_between_objects(o, player) or 10000
    local angleToPlayer = obj_angle_to_object(o, player) or 0
    if (distanceToPlayer < 1500 and player and math.abs(player.oPosY - o.oHomeY) < 500) then
      -- Point towards Mario, but only within 0x3000 angle units of the original angle. (original is 0x1500)
      local marioAngleFromWindSource = limit_angle(angleToPlayer - o.oSLSnowmanWindOriginalYaw)
      if (marioAngleFromWindSource > 0) then
        if (marioAngleFromWindSource < 0x3000) then
          o.oMoveAngleYaw = angleToPlayer
        else
          o.oMoveAngleYaw = o.oSLSnowmanWindOriginalYaw + 0x3000;
        end
      elseif (marioAngleFromWindSource > -0x3000) then
        o.oMoveAngleYaw = angleToPlayer
      else
        o.oMoveAngleYaw = o.oSLSnowmanWindOriginalYaw - 0x3000;
      end

      if o.oTimer > 180 then
        o.oTimer = 1
      elseif o.oTimer >= 90 then -- Spawn wind and play wind sound
        cur_obj_spawn_strong_wind_particles(12, 3, 0, 0, 0);
        cur_obj_play_sound_2(SOUND_AIR_BLOW_WIND)
      elseif o.oTimer >= 60 then -- breathe in
        if o.oTimer == 60 then
          cur_obj_play_sound_2(SOUND_OBJ_BOWSER_INHALING)
        end
        o.oMoveAngleYaw = o.oSLSnowmanWindOriginalYaw + 0x8000
        cur_obj_spawn_strong_wind_particles(12, 3, 0, 0, 0)
      end
    end
  else
    o.oTimer = 1
  end
end

hook_behavior(id_bhvSLSnowmanWind, OBJ_LIST_DEFAULT, true, custom_sl_wind_init, custom_sl_wind_loop, "bhvSLSnowmanWind")
hook_behavior(id_bhvSLWalkingPenguin, OBJ_LIST_SURFACE, true, obj_mark_for_deletion, nil, "bhvSLWalkingPenguin") -- no more walking penguin :(

-- dorrie
local DORRIE_ACT_ATTACK = DORRIE_ACT_RAISE_HEAD + 1
function custom_dorrie_loop(o)
  local m = nearest_mario_state_to_object(o)
  if o.oAction == DORRIE_ACT_ATTACK then
    if cur_obj_init_anim_check_frame(2, 35) ~= 0 then
      cur_obj_reverse_animation()
      if m.action & ACT_FLAG_INVULNERABLE == 0 and obj_is_near_to_and_facing_mario(m, 1000, 0x300) ~= 0 and dist_between_objects(o, m.marioObj) > 550 and m.waterLevel - m.pos.y < 250 then
        set_mario_action(m, ACT_EATEN_BY_BUBBA, 0)
      elseif m.action & ACT_FLAG_SWIMMING == 0 then
        o.oAction = DORRIE_ACT_RAISE_HEAD
      end
    else
      o.oDorrieNeckAngle = o.oDorrieNeckAngle + 0x115
    end
    obj_turn_toward_object(o, m.marioObj, 16, 0x100)
    o.oFaceAngleYaw = o.oMoveAngleYaw
  elseif o.oAction == DORRIE_ACT_MOVE and m.action & ACT_FLAG_SWIMMING ~= 0 then
    o.oAction = DORRIE_ACT_ATTACK
  end
end

hook_behavior(id_bhvDorrie, OBJ_LIST_SURFACE, false, nil, custom_dorrie_loop, "bhvDorrie") -- no more walking penguin :(

course_to_level = {
  [COURSE_NONE] = LEVEL_CASTLE_GROUNDS, -- Course 0 (note that courtyard and inside are also course 0); won't appear in minihunt
  [COURSE_BOB] = LEVEL_BOB,             -- Course 1
  [COURSE_WF] = LEVEL_WF,               -- Course 2
  [COURSE_JRB] = LEVEL_JRB,             -- Course 3
  [COURSE_CCM] = LEVEL_CCM,             -- Course 4
  [COURSE_BBH] = LEVEL_BBH,             -- Course 5
  [COURSE_HMC] = LEVEL_HMC,             -- Course 6
  [COURSE_LLL] = LEVEL_LLL,             -- Course 7
  [COURSE_SSL] = LEVEL_SSL,             -- Course 8
  [COURSE_DDD] = LEVEL_DDD,             -- Course 9
  [COURSE_SL] = LEVEL_SL,               -- Course 10
  [COURSE_WDW] = LEVEL_WDW,             -- Course 11
  [COURSE_TTM] = LEVEL_TTM,             -- Course 12
  [COURSE_THI] = LEVEL_THI,             -- Course 13
  [COURSE_TTC] = LEVEL_TTC,             -- Course 14
  [COURSE_RR] = LEVEL_RR,               -- Course 15
  [COURSE_BITDW] = LEVEL_BITDW,         -- Course 16 (also bowser 1)
  [COURSE_BITFS] = LEVEL_BITFS,         -- Course 17 (also bowser 2)
  [COURSE_BITS] = LEVEL_BITS,           -- Course 18 (also bowser 3)
  [COURSE_PSS] = LEVEL_PSS,             -- Course 19
  [COURSE_COTMC] = LEVEL_COTMC,         -- Course 20
  [COURSE_TOTWC] = LEVEL_TOTWC,         -- Course 21
  [COURSE_VCUTM] = LEVEL_VCUTM,         -- Course 22
  [COURSE_WMOTR] = LEVEL_WMOTR,         -- Course 23
  [COURSE_SA] = LEVEL_SA,               -- Course 24
  [25] = LEVEL_ENDING,                  -- Course 25 (will not appear in MiniHunt)
}
