-- handles packets and dark star saving/loading

gGlobalSyncTable.darkStarFlags = 0

-- save dark star
function save_dark_star(darkStarNum)
    if network_is_server() then
      local file = get_current_save_file_num() - 1
      local backup = (save_file_get_using_backup_slot() and 1) or 0
      local darkStarFlags = mod_storage_load("dark_"..file.."_"..backup)
      darkStarFlags = tonumber(darkStarFlags) or 0
      darkStarFlags = darkStarFlags | (1 << (darkStarNum - 1))
      mod_storage_save("dark_"..file.."_"..backup, tostring(darkStarFlags))
      gGlobalSyncTable.darkStarFlags = darkStarFlags
    else
      network_send_to(1, true, {
        id = PACKET_DARK_STAR,
        num = darkStarNum,
      })
    end
end

function load_dark_stars()
    if not (IS_VANILLA and network_is_server()) then return end
    local file = get_current_save_file_num() - 1
    local backup = (save_file_get_using_backup_slot() and 1) or 0
    
    if save_file_get_total_star_count(file, -1, COURSE_COUNT-1) == 0 then
        gGlobalSyncTable.darkStarFlags = 0
        mod_storage_save("dark_"..file.."_"..backup, "0")
        gGlobalSyncTable.secretBoss = false
        return
    end
    
    local darkStarFlags = mod_storage_load("dark_"..file.."_"..backup)
    darkStarFlags = tonumber(darkStarFlags) or 0
    gGlobalSyncTable.darkStarFlags = darkStarFlags
end

hook_event(HOOK_ON_LEVEL_INIT, load_dark_stars)

-- hud for dark stars in pause menu
local DARK_STAR_HUD = get_texture_info("dark_star_hud")
function on_hud_render()
    if (not (is_game_paused() and IS_VANILLA)) or djui_hud_is_pause_menu_created() then return end
    if gGlobalSyncTable.darkStarFlags == 0 then return end
    
    djui_hud_set_resolution(RESOLUTION_N64)
    local screenWidth = djui_hud_get_screen_width()
    local screenHeight = djui_hud_get_screen_height()
    local x = (screenWidth - 18 * 3 + 2) * 0.5
    local y = screenHeight - 20
    for i=0,2 do
        if gGlobalSyncTable.darkStarFlags & (1 << i) ~= 0 then
            djui_hud_set_color(255, 255, 255, 255)
        else
            djui_hud_set_color(0, 0, 0, 100)
        end
        djui_hud_render_texture(DARK_STAR_HUD, x, y, 1, 1)
        x = x + 18
    end
end
hook_event(HOOK_ON_HUD_RENDER, on_hud_render)

function on_packet_dark_star(data)
    save_dark_star(data.num)
end

function on_packet_update_stars()
    update_all_mario_stars()
end

function on_packet_hundred_lives()
    gMarioStates[0].numLives = 100
    cur_obj_play_sound_2(SOUND_GENERAL_COLLECT_1UP)
end

PACKET_DARK_STAR = 0
PACKET_UPDATE_STARS = 1
PACKET_HUNDRED_LIVES = 2
local sPacketList = {
    [PACKET_DARK_STAR] = on_packet_dark_star,
    [PACKET_UPDATE_STARS] = on_packet_update_stars,
    [PACKET_HUNDRED_LIVES] = on_packet_hundred_lives,
}
function on_packet_recieve(data)
    if sPacketList[data.id] then
        sPacketList[data.id](data)
    end
end
hook_event(HOOK_ON_PACKET_RECEIVE, on_packet_recieve)