-- yoshi fight

local eggSpawnObjects = {
    { id_bhvBobomb,           E_MODEL_BLACK_BOBOMB },
    { id_bhvGoomba,           E_MODEL_GOOMBA },
    { id_bhvFlyGuy,           E_MODEL_FLYGUY },
    { id_bhvSnufit,           E_MODEL_SNUFIT },
    { id_bhvScuttlebug,       E_MODEL_SCUTTLEBUG },
    { id_bhvFirePiranhaPlant, E_MODEL_PIRANHA_PLANT },
    { id_bhvBoo,              E_MODEL_BOO },
    { id_bhvKoopa,            E_MODEL_KOOPA_WITHOUT_SHELL },
}

local YOSHI_ACT_JUMP_FLUTTER = YOSHI_ACT_CREDITS + 1
local YOSHI_ACT_GROUND_POUND = YOSHI_ACT_CREDITS + 2
local YOSHI_ACT_THROW_EGGS = YOSHI_ACT_CREDITS + 3
local YOSHI_ACT_DAMAGED = YOSHI_ACT_CREDITS + 4

---@param o Object
function custom_yoshi_init(o)
    o.oFlags = (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
    o.oAnimations = gObjectAnimations.yoshi_seg5_anims_05024100
    o.oInteractType = INTERACT_TEXT
    cur_obj_update_floor()
    o.oIntangibleTimer = 0
    if IS_VANILLA then
        cur_obj_set_hitbox_radius_and_height(100, 150)
    else
        cur_obj_set_hitbox_radius_and_height(160, 150)
    end
    cur_obj_init_animation(0)
    o.oHomeX, o.oHomeY, o.oHomeZ = o.oPosX, o.oPosY, o.oPosZ
    bhv_yoshi_init()
    if IS_VANILLA then network_init_object(o, true, {'oSubAction', 'oHealth'}) end
end

---@param o Object
function custom_yoshi_loop(o)
    if not IS_VANILLA then
        bhv_yoshi_loop()
        o.oIntangibleTimer = 0
        return
    end

    if o.oAction == YOSHI_ACT_TALK then
        local charName = gMarioStates[0].character.name
        local dialog = ""
        if charSelectExists then
            local charData = charSelect.character_get_current_table()
            if charData and charData.name and not (charData.name:find(charName) and charData.name:find(" ")) then
                charName = charData.name
            end
        end
        if charName == "Bowser" or charName == "Bowser Jr" or charName == "Bowsette" then
            dialog = string.format(DIALOG_YOSHI_BOWSER, charName)
        elseif charName:find("Yoshi") then
            dialog = DIALOG_YOSHI_SELF
        else
            dialog = string.format(DIALOG_YOSHI_FIGHT, charName)
        end
        smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.YoshiDialog, 1, 4, 30, 200, dialog)
        playIntroOverride = true
        bhv_yoshi_loop()
        o.oIntangibleTimer = 0
        return
    elseif o.oAction == YOSHI_ACT_WALK and playIntroOverride then
        cur_obj_change_action(YOSHI_ACT_JUMP_FLUTTER)
        o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 40, -2, -0.5, 10, 10, 16
        o.oInteractType = INTERACT_DAMAGE

        local sYoshiHitbox = get_temp_object_hitbox()
        sYoshiHitbox.interactType = INTERACT_DAMAGE
        sYoshiHitbox.downOffset = 0
        sYoshiHitbox.damageOrCoinValue = 1
        sYoshiHitbox.health = 3
        sYoshiHitbox.numLootCoins = 0
        sYoshiHitbox.radius = 100
        sYoshiHitbox.height = 150
        sYoshiHitbox.hurtboxRadius = 100
        sYoshiHitbox.hurtboxHeight = 150
        obj_set_hitbox(o, sYoshiHitbox)

        o.oHealth = 3
        playIntroOverride = false
        play_music(SEQ_PLAYER_LEVEL, SEQ_EVENT_BOSS, 0)
        return
    elseif o.oAction <= YOSHI_ACT_CREDITS then
        bhv_yoshi_loop()
        o.oIntangibleTimer = 0
        return
    end

    -- fight code
    o.oAnimState = 0
    local m = nearest_mario_state_to_object(o)
    if o.oInteractType == INTERACT_TEXT then
        local oldHealth = o.oHealth
        o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 40, -2, -0.5, 10, 10, 16
        o.oInteractType = INTERACT_DAMAGE

        local sYoshiHitbox = get_temp_object_hitbox()
        sYoshiHitbox.interactType = INTERACT_DAMAGE
        sYoshiHitbox.downOffset = 0
        sYoshiHitbox.damageOrCoinValue = 1
        sYoshiHitbox.health = 3
        sYoshiHitbox.numLootCoins = 0
        sYoshiHitbox.radius = 100
        sYoshiHitbox.height = 150
        sYoshiHitbox.hurtboxRadius = 100
        sYoshiHitbox.hurtboxHeight = 150
        obj_set_hitbox(o, sYoshiHitbox)

        o.oHealth = oldHealth
        playIntroOverride = false
        play_music(SEQ_PLAYER_LEVEL, SEQ_EVENT_BOSS, 0)
    end

    if m.health <= 0xFF then
        local allDead = true
        for i = 0, MAX_PLAYERS - 1 do
            local m2 = gMarioStates[i]
            if is_player_active(m2) ~= 0 and m2.health > 0xFF then
                allDead = false
                break
            end
        end
        if allDead then
            for i, obj in ipairs(eggSpawnObjects) do
                local other = obj_get_first_with_behavior_id(obj[1])
                while other do
                    if obj[1] == id_bhvBoo then
                        other.oInteractStatus = INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
                    else
                        obj_mark_for_deletion(other)
                    end
                    other = obj_get_next_with_same_behavior_id(other)
                end
            end

            local other = obj_get_first_with_behavior_id(id_bhvYoshiEgg)
            while other do
                obj_mark_for_deletion(other)
                other = obj_get_next_with_same_behavior_id(other)
            end

            cur_obj_set_pos_to_home()
            o.oInteractType = INTERACT_TEXT
            cur_obj_init_animation(0)
            cur_obj_change_action(YOSHI_ACT_IDLE)
            o.oIntangibleTimer = 0
            o.oInteractStatus = 0
            bhv_yoshi_init()
        end
    end

    if o.oAction == YOSHI_ACT_JUMP_FLUTTER then
        local angleToPlayer = obj_angle_to_object(o, m.marioObj)
        local distanceToPlayer = lateral_dist_between_objects(o, m.marioObj)

        if (o.oSubAction == 0 and o.oTimer == 1) or (o.oMoveFlags & (OBJ_MOVE_LANDED | OBJ_MOVE_MASK_IN_WATER)) ~= 0 then
            o.oVelY = 40
            o.oSubAction = 0
            o.oMoveAngleYaw = angleToPlayer
            cur_obj_play_sound_2(SOUND_GENERAL_ENEMY_ALERT1)
        end

        if o.oSubAction == 0 then
            cur_obj_init_animation(2)
            cur_obj_extend_animation_if_at_end()
            if o.oVelY < -10 and o.oPosY - o.oHomeY < 500 then
                o.oSubAction = 1
                o.oVelY = -10
                o.oTimer = 0
                cur_obj_play_sound_2(SOUND_GENERAL_YOSHI_TALK)
            end
        else
            if o.oTimer < 25 then
                o.oVelY = o.oVelY + 4
                cur_obj_init_animation_with_accel_and_sound(1, 5)
            else
                o.oSubAction = 0
                o.oTimer = 1
            end
            cur_obj_play_sound_2(SOUND_ACTION_TWIRL)
        end

        cur_obj_rotate_yaw_toward(angleToPlayer, 0x800)
        o.oForwardVel = clamp(10, distanceToPlayer // 20, 32)
        o.oFaceAngleYaw = o.oMoveAngleYaw
        if (o.oVelY < 0 or o.oSubAction == 1) and distanceToPlayer < 200 then
            cur_obj_change_action(YOSHI_ACT_GROUND_POUND)
            o.oSubAction = 0
        end
    elseif o.oAction == YOSHI_ACT_GROUND_POUND then
        cur_obj_init_animation_and_anim_frame(2, 5)
        o.oVelY = -60
        o.oForwardVel = 0
        o.oDamageOrCoinValue = 4
        if ((o.oMoveFlags & (OBJ_MOVE_LANDED | OBJ_MOVE_MASK_IN_WATER)) ~= 0 or o.oTimer > 300) and o.oSubAction == 0 then
            o.oSubAction = 1
            o.oTimer = 0
            cur_obj_play_sound_2(SOUND_ACTION_TERRAIN_HEAVY_LANDING)
            set_camera_shake_from_point(SHAKE_POS_SMALL, o.oPosX, o.oPosY, o.oPosZ)
            spawn_mist_particles()
            o.hurtboxRadius = 300
        elseif o.oSubAction == 1 then
            o.oDamageOrCoinValue = 1
            o.hurtboxRadius = 100
            if o.oTimer > 10 then
                cur_obj_change_action(YOSHI_ACT_THROW_EGGS)
                o.oVelY = 0
                o.oSubAction = 0
            end
        end
    elseif o.oAction == YOSHI_ACT_THROW_EGGS then
        local angleToPlayer = obj_angle_to_object(o, m.marioObj)
        cur_obj_rotate_yaw_toward(angleToPlayer, 0x800)
        cur_obj_init_animation(2)

        o.oForwardVel = 0
        o.oVelY = 0
        if o.oTimer >= 20 then
            if o.oSubAction > 5 - o.oHealth then
                cur_obj_change_action(YOSHI_ACT_JUMP_FLUTTER)
                o.oSubAction = 0
            else
                o.oTimer = 0
                o.oSubAction = o.oSubAction + 1
                play_sound_with_freq_scale(SOUND_GENERAL_YOSHI_TALK, o.header.gfx.cameraToObject, 2)
                if m.playerIndex == 0 then
                    spawn_sync_object(id_bhvYoshiEgg, E_MODEL_YOSHI_EGG, o.oPosX, o.oPosY, o.oPosZ, function(egg)
                        egg.oMoveAngleYaw = o.oMoveAngleYaw
                        egg.oBehParams = (6 - o.oHealth) - o.oSubAction
                        egg.oVelY = 40
                        o.oForwardVel = 30
                    end)
                end
            end
        else
            cur_obj_extend_animation_if_at_end()
        end
    elseif o.oAction == YOSHI_ACT_DAMAGED then
        cur_obj_init_animation(0)
        o.oAnimState = 1
        o.oForwardVel = 0
        if o.oTimer == 1 then
            play_sound_with_freq_scale(SOUND_GENERAL_YOSHI_TALK, o.header.gfx.cameraToObject, 0.75)
            o.oHealth = o.oHealth - 1
        end

        if o.oTimer < 60 then
            o.oMoveAngleYaw = o.oMoveAngleYaw + 0x1000
        elseif o.oTimer > 75 then
            if o.oHealth > 0 then
                cur_obj_change_action(YOSHI_ACT_JUMP_FLUTTER)
                o.oSubAction = 0
            elseif m and m.playerIndex == 0 and should_start_or_continue_dialog(m, o) ~= 0 then
                if m.action ~= ACT_READING_NPC_DIALOG then
                    set_mario_action(m, ACT_READING_NPC_DIALOG, 0)
                    smlua_text_utils_dialog_replace(gBehaviorValues.dialogs.YoshiDialog, 1, 4, 30, 200,
                        DIALOG_YOSHI_DEFEAT)
                end

                for i, obj in ipairs(eggSpawnObjects) do
                    local other = obj_get_first_with_behavior_id(obj[1])
                    while other do
                        if obj[1] == id_bhvBoo then
                            other.oInteractStatus = INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED
                        else
                            obj_mark_for_deletion(other)
                        end
                        other = obj_get_next_with_same_behavior_id(other)
                    end
                end

                local other = obj_get_first_with_behavior_id(id_bhvYoshiEgg)
                while other do
                    obj_mark_for_deletion(other)
                    other = obj_get_next_with_same_behavior_id(other)
                end

                if cutscene_object_with_dialog(CUTSCENE_DIALOG, o, gBehaviorValues.dialogs.YoshiDialog) ~= 0 then
                    m.numLives = 100
                    cur_obj_play_sound_2(SOUND_GENERAL_COLLECT_1UP)
                    network_send(true, { id = PACKET_HUNDRED_LIVES })

                    spawn_mist_particles()
                    stop_background_music(SEQ_EVENT_BOSS)
                    seq_player_unlower_volume(SEQ_PLAYER_LEVEL, 60)
                    o.oBehParams = (10 << 24) + 3
                    spawn_no_exit_star(o.oHomeX, o.oHomeY + 300, o.oHomeZ)
                    if o.activeFlags & ACTIVE_FLAG_DEACTIVATED == 0 then
                        obj_mark_for_deletion(o)
                        network_send_object(o, true)
                    end
                end
            end
        end
    end

    if o.oAction ~= YOSHI_ACT_DAMAGED and o.oInteractStatus & INT_STATUS_TOUCHED_BOB_OMB ~= 0 then
        cur_obj_change_action(YOSHI_ACT_DAMAGED)
        network_send_object(o, true)
    end

    o.oInteractStatus = 0
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)
end

id_bhvYoshi = hook_behavior(id_bhvYoshi, OBJ_LIST_GENACTOR, true, custom_yoshi_init, custom_yoshi_loop, "bhvYoshi")

-- yoshi egg

function yoshi_egg_init(o)
    o.oFlags = (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
    o.oInteractType = INTERACT_DAMAGE

    local sYoshiEggHitbox = get_temp_object_hitbox()
    sYoshiEggHitbox.interactType = INTERACT_DAMAGE
    sYoshiEggHitbox.health = 3
    sYoshiEggHitbox.downOffset = 0
    sYoshiEggHitbox.damageOrCoinValue = 2
    sYoshiEggHitbox.numLootCoins = 0
    sYoshiEggHitbox.radius = 100
    sYoshiEggHitbox.height = 100
    sYoshiEggHitbox.hurtboxRadius = 100
    sYoshiEggHitbox.hurtboxHeight = 100
    obj_set_hitbox(o, sYoshiEggHitbox)

    obj_set_billboard(o)
    cur_obj_scale(2)
    o.oIntangibleTimer = 0
    o.oWallHitboxRadius, o.oGravity, o.oBounciness, o.oDragStrength, o.oFriction, o.oBuoyancy = 40, -4, 0, 10, 10, 16
    o.oAnimState = 0
    o.oSubAction = 0
    network_init_object(o, true, {})
end

function yoshi_egg_loop(o)
    local m = nearest_mario_state_to_object(o)
    if (o.oMoveFlags & (OBJ_MOVE_MASK_ON_GROUND | OBJ_MOVE_MASK_IN_WATER)) ~= 0 then
        o.oForwardVel = 0
        o.oAnimState = o.oAnimState + 1

        if o.oAnimState >= 7 then
            o.oSubAction = o.oSubAction + 1
            if o.oSubAction >= 5 and (o.oBehParams ~= 0 or o.oSubAction >= 7) then
                spawn_triangle_break_particles(20, 138, 0.7, 3)
                cur_obj_play_sound_2(SOUND_GENERAL_BREAK_BOX)
                local randomObjData = eggSpawnObjects[1]
                if o.oBehParams ~= 0 then
                    randomObjData = eggSpawnObjects[math.random(2, #eggSpawnObjects)]
                end
                if m.playerIndex == 0 then
                    spawn_sync_object(randomObjData[1], randomObjData[2], o.oPosX, o.oPosY + 300, o.oPosZ,
                        function(other)
                            other.oBehParams = 0x100
                            other.oNumLootCoins = 0
                            other.oMoveAngleYaw = o.oFaceAngleYaw
                            other.oMoveAnglePitch = 0
                            if randomObjData[1] == id_bhvKoopa then
                                other.oKoopaMovementType = KOOPA_BP_UNSHELLED
                            end
                        end)
                end
                obj_mark_for_deletion(o)
                return
            end
            local angleToPlayer = obj_angle_to_object(o, m.marioObj)
            cur_obj_rotate_yaw_toward(angleToPlayer, 0x4000)
            o.oVelY = 40
            o.oAnimState = 0
            cur_obj_play_sound_2(SOUND_OBJ_WATER_BOMB_BOUNCING)
        end
    else
        o.oTimer = 0
        o.oForwardVel = 30
    end

    o.oInteractStatus = 0
    cur_obj_update_floor_and_walls()
    cur_obj_move_standard(-78)
end

-- default so that it doesnt trigger collision
id_bhvYoshiEgg = hook_behavior(nil, OBJ_LIST_DEFAULT, true, yoshi_egg_init, yoshi_egg_loop, "bhvYoshiEgg")
